# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: pymca
   :platform: Unix
   :synopsis: A plugin to fit spectral data

.. moduleauthor:: Aaron D. Parsons <scientificsoftware@diamond.ac.uk>
"""

import logging
from savu.plugins.filters.base_filter import BaseFilter
from savu.plugins.driver.cpu_plugin import CpuPlugin
from savu.plugins.utils import register_plugin, dawn_compatible, OUTPUT_TYPE_METADATA_ONLY
import numpy as np
import os
import savu.test.test_utils as tu
from PyMca5.PyMcaPhysics.xrf import McaAdvancedFitBatch


@dawn_compatible(OUTPUT_TYPE_METADATA_ONLY)
@register_plugin
class Pymca(BaseFilter, CpuPlugin):
    """
    uses pymca to fit spectral data

    :u*param config: path to the config file. Default: 'Savu/test_data/data/test_config.cfg'.

    """

    def __init__(self):
        logging.debug("fitting spectrum")
        super(Pymca, self).__init__("Pymca")

    def pre_process(self):
        in_dataset, _out_datasets = self.get_datasets()
        in_d1 = in_dataset[0]
        self.sh = in_d1.get_shape()
        self.b = self.setup_fit(np.random.random((1,1,self.sh[-1])))

    def process_frames(self, data):
        y = np.expand_dims(data,0)
        self.b = self.setup_fit(y)
        self.b._McaAdvancedFitBatch__processStack()
        try:
            stack = self.b.outbuffer['parameters']
            op_stack = np.rollaxis(stack,0,3)
        except (AttributeError, KeyError) as e:
            op_stack = -np.ones((1,1,self.outputshape[-1]))
            logging.warning("Error in fit:%s",e)
        op = op_stack[0,0]
        return op

    def setup(self):
        logging.debug('setting up the pymca fitting')
        in_dataset, out_datasets = self.get_datasets()
        inshape = in_dataset[0].get_shape()
        spectra_shape = inshape[-1] #  nearly always true.
        rest_shape = inshape[:-1]
        np.random.seed=1
        dummy_spectrum = np.random.random((1, 1, spectra_shape))
        c = self.setup_fit(dummy_spectrum)# seed it with junk, zeros made the matrix singular unsurprisingly and this bungles it.

        # temporary measure to stop the next line printing arrays to screen.
        c.processList()#_McaAdvancedFitBatch__processStack()# perform an initial fit to get the shapes

        fit_labels = c.outbuffer.labels('parameters') # and then take out the axis labels for the channels
        out_meta_data = out_datasets[0].meta_data
        out_meta_data.set("PeakElements",fit_labels)

        self.outputshape = rest_shape+(len(fit_labels),) # and this is the shape the thing will be
#         print "input shape is", in_dataset[0].get_shape()
#         print "the output shape in setup is"+str(outputshape)

        axis_labels = ['-1.PeakElements.label']
        in_patterns = in_dataset[0].get_data_patterns()
#         pattern_list = ['SINOGRAM', 'PROJECTION']
        pattern_list = list(in_patterns.keys())

        fitResult = out_datasets[0]

        fitResult.create_dataset(patterns={in_dataset[0]: pattern_list},
                                axis_labels={in_dataset[0]: axis_labels},
                                shape=self.outputshape)
        fitResult.meta_data.set('FitAreas',np.array(fit_labels))
        slice_directions = tuple(range(len(rest_shape)))
#         print "slice directions are:"+str(slice_directions)
        fitResult.add_pattern("CHANNEL", core_dims=(-1,),
                        slice_dims=slice_directions)

        in_pData, out_pData = self.get_plugin_datasets()
        in_pData[0].plugin_data_setup(self.get_plugin_pattern(), self.get_max_frames())
        out_pData[0].plugin_data_setup('CHANNEL', self.get_max_frames())


    def get_max_frames(self):
        return 'single'

    def get_plugin_pattern(self):
        return 'SPECTRUM'

    def nOutput_datasets(self):
        return 1

    def setup_fit(self,y):
        '''
        takes a data shape and returns a fit-primed object
        '''
        outputdir=None # nope
        roifit=0# nope
        roiwidth=y.shape[1] #need this to pretend its an image
        b = McaAdvancedFitBatch.McaAdvancedFitBatch(self.get_conf_path(),
                                                    [y],
                                                    outputdir,
                                                    roifit,
                                                    roiwidth,
                                                    fitfiles=0,
                                                    nosave=True,
                                                    quiet=True) # prime the beauty
        b.pleaseBreak = 1

        # temporary measure to stop the next line printing arrays to screen.
        b.processList()

        b.pleaseBreak = 0
        return b

    def get_conf_path(self):
        path = self.parameters['config']
        if path.split(os.sep)[0] == 'Savu':
            path = tu.get_test_data_path(path.split('/test_data/data')[1])
        return path

    def get_dummyhdf_path(self):
        return tu.get_test_data_path('i18_test_data.nxs')



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: yaml_utils
   :platform: Unix
   :synopsis: Utilities for yaml files

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""

import yaml
from collections import OrderedDict


def ordered_load(stream, Loader=yaml.Loader, object_pairs_hook=OrderedDict):
    class OrderedLoader(Loader):
        pass

    def construct_mapping(loader, node):
        loader.flatten_mapping(node)
        return object_pairs_hook(loader.construct_pairs(node))
    OrderedLoader.add_constructor(
        yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
        construct_mapping)
    return yaml.load(stream, OrderedLoader)


def ordered_dump(data, stream=None, Dumper=yaml.Dumper, **kwds):
    class OrderedDumper(Dumper):
        pass

    def _dict_representer(dumper, data):
        return dumper.represent_mapping(
            yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
            list(data.items()))
    OrderedDumper.add_representer(OrderedDict, _dict_representer)
    
    def represent_none(self, _):
        return self.represent_scalar('tag:yaml.org,2002:null', '')
    OrderedDumper.add_representer(type(None), represent_none)
    
    return yaml.dump(data, stream, OrderedDumper, **kwds)


def read_yaml(path):
    with open(path, 'r') as stream:
        data_dict = ordered_load(stream, yaml.SafeLoader)
    return data_dict


def dump_yaml(template, stream):
    ordered_dump(template, stream=stream, Dumper=yaml.SafeDumper,
                 default_flow_style=False)




# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
.. module:: utils
   :platform: Unix
   :synopsis: Utilities for plugin management

.. moduleauthor:: Mark Basham <scientificsoftware@diamond.ac.uk>

"""

import os
import re
import sys
import ast
import logging
import savu
import importlib
import inspect
import itertools

from collections import OrderedDict

# can I remove these from here?
plugins = {}
plugins_path = {}
dawn_plugins = {}
count = 0

OUTPUT_TYPE_DATA_ONLY = 0
OUTPUT_TYPE_METADATA_ONLY = 1
OUTPUT_TYPE_METADATA_AND_DATA = 2


def register_plugin(clazz):
    """decorator to add plugins to a central register"""
    plugins[clazz.__name__] = clazz
    if clazz.__module__.split('.')[0] != 'savu':
        plugins_path[clazz.__name__] = clazz.__module__
    return clazz


def dawn_compatible(plugin_output_type=OUTPUT_TYPE_METADATA_AND_DATA):
    def _dawn_compatible(clazz):
        """
        decorator to add dawn compatible plugins and details to a central
        register
        """
        dawn_plugins[clazz.__name__] = {}
        try:
            plugin_path = sys.modules[clazz.__module__].__file__
            # looks out for .pyc files
            dawn_plugins[clazz.__name__]['path2plugin'] = plugin_path.split('.py')[0] + '.py'
            dawn_plugins[clazz.__name__]['plugin_output_type'] = _plugin_output_type
        except Exception as e:
            print(e)
        return clazz

    # for backwards compatibility, if decorator is invoked without brackets...
    if inspect.isclass(plugin_output_type):
        _plugin_output_type = OUTPUT_TYPE_METADATA_AND_DATA
        return _dawn_compatible(plugin_output_type)
    else:
        _plugin_output_type = plugin_output_type
        return _dawn_compatible


def get_plugin(plugin_name, params, exp, check=False):
    """ Get an instance of the plugin class and populate default parameters.

    :param plugin_name: Name of the plugin to import
    :type plugin_name: str.
    :returns:  An instance of the class described by the named plugin.
    """
    logging.debug("Importing the module %s", plugin_name)
    instance = load_class(plugin_name)()
    instance.initialise(params, exp, check=check)
    return instance


def _get_cls_name(name):
    return ''.join(x.capitalize() for x in name.split('.')[-1].split('_'))


def load_class(name, cls_name=None):
    """ Returns an instance of the class associated with the module name.

    :param name: Module name or path to a module file
    :returns: An instance of the class associated with module.
    """
    path = name if os.path.dirname(name) else None
    name = os.path.basename(os.path.splitext(name)[0]) if path else name
    cls_name = _get_cls_name(name) if not cls_name else cls_name
    if cls_name in plugins.keys():
        return plugins[cls_name]
    if path:
        mod = importlib.machinery.SourceFileLoader(name, path).load_module()
    else:
        mod = importlib.import_module(name)
    return getattr(mod, cls_name)


def plugin_loader(exp, plugin_dict, check=False):
    logging.debug("Running plugin loader")
    try:
        plugin = get_plugin(plugin_dict['id'],
                            plugin_dict['data'],
                            exp,
                            check=check)
    except Exception as e:
        logging.error("failed to load the plugin")
        logging.error(e)
        # re-raise the original error
        raise

    if check:
        exp.meta_data.plugin_list._set_datasets_list(plugin)

    logging.debug("finished plugin loader")
    return plugin


def get_plugins_paths(examples=True):
    """
    This gets the plugin paths, but also adds any that are not on the
    pythonpath to it.
    """
    plugins_paths = OrderedDict()
    
    # Add the savu plugins paths first so it is overridden by user folders
    savu_plugins_path = os.path.join(savu.__path__[0], 'plugins')
    savu_plugins_subpaths = [d for d in next(os.walk(savu_plugins_path))[1] \
                             if d != "__pycache__"]
    for path in savu_plugins_subpaths:
        plugins_paths[os.path.join(savu_plugins_path, path)] = \
            ''.join(['savu.plugins.', path, '.'])
    
    # get user, environment and example plugin paths
    user_path = [os.path.join(os.path.expanduser("~"), 'savu_plugins')]
    env_paths = os.getenv("SAVU_PLUGINS_PATH", "").replace(" ", "").split(":")
    templates = "../plugin_examples/plugin_templates"
    eg_path = [os.path.join(savu.__path__[0], templates)] if examples else []

    for ppath in env_paths + user_path + eg_path:
        if os.path.exists(ppath):
            plugins_paths[ppath] = os.path.basename(ppath) + '.'
            if ppath not in sys.path:
                sys.path.append(os.path.dirname(ppath))

    return plugins_paths


def is_template_param(param):
    """ Identifies if the parameter should be included in an input template
    and returns the default value of the parameter if it exists.
    """
    start = 0
    ptype = 'local'
    if isinstance(param, str):
        param = param.strip()
        if not param.split('global')[0]:
            ptype = 'global'
            start = 6
        first, last = param[start], param[-1]
        if first == '<' and last == '>':
            param = param[start + 1:-1]
            param = None if not param else param
            try:
                param = eval(param)
            except:
                pass
            return [ptype, param]
    return False


def blockPrint():
    """ Disable printing to stdout """
    import tempfile
    fname = tempfile.mkdtemp() + '/unwanted_prints.txt'
    sys.stdout = open(fname, 'w')


def enablePrint():
    """ Enable printing to stdout """
    sys.stdout = sys.__stdout__


def parse_config_string(string):
    regex = r"[\[\]\, ]+"
    split_vals = [_f for _f in re.split(regex, string) if _f]
    delimitors = re.findall(regex, string)
    split_vals = [repr(a.strip()) for a in split_vals]
    zipped = itertools.zip_longest(delimitors, split_vals)
    string = ''.join([i for l in zipped for i in l if i is not None])
    try:
        return ast.literal_eval(string)
    except ValueError:
        return ast.literal_eval(parse_array_index_as_string(string))


def parse_array_index_as_string(string):
    p = re.compile(r"'\['")
    for m in p.finditer(string):
        offset = m.start() - count + 3
        end = string[offset:].index("']") + offset
        string = string[:end] + "]'" + string[end + 2:]
    string = string.replace("'['", '[')
    return string



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: Large field of view loader
   :platform: Unix
   :synopsis: A class for loading multiple standard tomography scans.

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""

import copy
import numpy as np
import glob

from savu.plugins.utils import register_plugin
from savu.plugins.loaders.full_field_loaders.multi_nxtomo_loader import \
        MultiNxtomoLoader


@register_plugin
class LfovLoader(MultiNxtomoLoader):
    """
    A class to load 2 scans in Nexus/hdf format into one dataset.

    :u*param file_name: The shared part of the name of each file\
        (not including .nxs). Default: 'projection'.
    :param data_path: Path to the data inside the \
        file. Default: 'entry/data/data'.
    :param order: Order of datasets used for stitching. Default: [1, 0].
    :param row_offset: Offsets of row indices between datasets. Default: [0, -1].

    :*param stack_or_cat: Stack or concatenate the data\
        (4D and 3D respectively). Default: 'stack'.
    :*param stack_or_cat_dim: Dimension to stack or concatenate. Default: 3.
    :*param axis_label: New axis label, if required, in the form\
        'name.units'. Default: 'scan.number'.
    """

    def __init__(self, name='LfovLoader'):
        super(LfovLoader, self).__init__(name)
        
    def _update_preview(self, preview, offset):
        """
        Add extra sinograms to the preview if there are not enough to stitch
        at least one frame after taking into account the offset between images.

        Parameters
        ----------
        preview : list
            A Savu preview data list.

        Returns
        -------
        None.

        """
        if not offset or not preview:
            return preview
        dObj = self.exp.index['in_data']['tomo']

        # Revert the data shape to before previewing was applied
        dObj.set_shape(dObj.get_original_shape())
        preview = dObj.get_preview().get_integer_entries(preview)
        sino_slice_dim = dObj.get_data_dimension_by_axis_label("detector_y")
        max_sino_idx = dObj.data.shape[sino_slice_dim]
        sl = list(map(int, preview[sino_slice_dim].split(":")))
        sl_0 = sl[0] + offset
        sl_1 = sl[1] + offset
        if (max_sino_idx >= sl_0 >= 0) and (max_sino_idx >= sl_1 >= 0):
            sl[0] = sl_0
            sl[1] = sl_1   
            preview[sino_slice_dim] = ":".join(map(str, sl))
            self.parameters['preview'][sino_slice_dim] = preview[sino_slice_dim]
        return preview

    def _get_order_list(self):
        order = self.parameters['order']
        if order[0] < order[-1]:
            order_list = np.arange(order[0], order[1] + 1)
        else:
            order_list = np.arange(order[0], order[1] - 1, -1)        
        return order_list

    def _get_file_list(self, order_list):
        shared_name = self.parameters["file_name"]
        file_path = copy.copy(self.exp.meta_data.get('data_file'))
        self.exp.meta_data.set('data_file', file_path)
        
        if shared_name is None:
            file_list = self._find_files(file_path, "/*.hdf")
            if len(file_list) == 0:
                file_list = self._find_files(file_path, "/*.nxs")
        else:
            file_list = self._find_files(file_path, "/*" + shared_name + "*")

        if len(order_list) != len(file_list):
            raise ValueError(
                "Number of files found in the folder is not the same as the"
                +" requested number")
        return file_list
    
    def _find_files(self, base, name):
        return sorted(glob.glob(base + name)) if base else []

    def _get_data_objects(self, nxtomo):
        order_list = self._get_order_list()
        file_list = self._get_file_list(order_list)        
        dark_folder, dark_key, dark_scale = self.parameters['dark']
        flat_folder, flat_key, flat_scale = self.parameters['flat']
        dark_list = self._find_files(dark_folder, "/*dark*")
        flat_list = self._find_files(flat_folder, "/*flat*")
        offset = self.parameters['row_offset']
       
        data_obj_list = []
        for i in order_list:
            self.exp.meta_data.set('data_file', file_list[i])
            # update darks and flats
            if dark_folder is not None:
                nxtomo.parameters['dark'] = [
                    dark_list[i], dark_key, dark_scale]
            if flat_folder is not None:
                nxtomo.parameters['flat'] = [
                    flat_list[i], flat_key, flat_scale]
            nxtomo.setup()
            # update preview
            if offset[i] != 0:
                nxtomo.parameters['preview'] = \
                    self._update_preview(nxtomo.parameters['preview'], offset[i])
            data_obj_list.append(self.exp.index['in_data']['tomo'])
            self.exp.index['in_data'] = {}
        return data_obj_list



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: tomo_recon
   :platform: Unix
   :synopsis: Runner for the Savu framework

.. moduleauthor:: Mark Basham <scientificsoftware@diamond.ac.uk>

"""

import tempfile  # this import is required for pyFAI - DO NOT REMOVE!
import argparse
import traceback
import sys
import os
from mpi4py import MPI
from savu.version import __version__

from savu.core.basic_plugin_runner import BasicPluginRunner
from savu.core.plugin_runner import PluginRunner


def __option_parser():
    """ Option parser for command line arguments.
    """
    version = "%(prog)s " + __version__
    parser = argparse.ArgumentParser(prog='savu')
    hide = argparse.SUPPRESS

    parser.add_argument('in_file', help='Input data file.')
    process_str = 'Process list, created with the savu configurator.'
    parser.add_argument('process_list', help=process_str)
    parser.add_argument('out_folder', help='Output folder.')
    parser.add_argument('--version', action='version', version=version)
    parser.add_argument("-f", "--folder", help="Override output folder name")

    tmp_help = "Store intermediate files in a temp directory."
    parser.add_argument("-d", "--tmp", help=tmp_help)

    template_help = "Pass a template file of plugin input parameters."
    parser.add_argument("-t", "--template", help=template_help, default=None)

    log_help = "Store full log file in a separate location"
    parser.add_argument("-l", "--log", help=log_help)

    v_help = "Display all debug log messages"
    parser.add_argument("-v", "--verbose", help=v_help, action="store_true",
                        default=False)
    parser.add_argument("-q", "--quiet", action="store_true", dest="quiet",
                        help="Display only Errors and Info.", default=False)
    # temporary flag to fix lustre issue
    parser.add_argument("--lustre_workaround", action="store_true",
                        dest="lustre", help="Avoid lustre segmentation fault",
                        default=False)
    sys_params_help = "Override default path to Savu system parameters file."
    parser.add_argument("--system_params", help=sys_params_help, default=None)

    # Hidden arguments
    # process names
    parser.add_argument("-n", "--names", help=hide, default="CPU0")
    # transport mechanism
    parser.add_argument("--transport", help=hide, default="hdf5")
    # Set Savu mode
    parser.add_argument("-m", "--mode", help=hide, default="full",
                        choices=['basic', 'full'])
    # Set logging to cluster mode
    parser.add_argument("-c", "--cluster", action="store_true", help=hide,
                        default=False)
    # Send an email on completion
    parser.add_argument("-e", "--email", dest="email", help=hide, default=None)
    # Facility email for errors
    parser.add_argument("--facility_email", dest="femail", help=hide,
                        default=None)
    # Set beamline log file (for online processing)
    parser.add_argument("--bllog", dest="bllog", help=hide, default=None)
    # Location of syslog server
    parser.add_argument("-s", "--syslog", dest="syslog", help=hide,
                        default='localhost')
    # Port to connect to syslog server on
    parser.add_argument("-p", "--syslog_port", dest="syslog_port",
                        help=hide, default=514, type=int)
    parser.add_argument("--test_state", dest="test_state", default='False',
                        action='store_true', help=hide)

    # DosNa related parameters
    parser.add_argument("--dosna_backend", dest="dosna_backend", help=hide,
                        default=None)
    parser.add_argument("--dosna_engine", dest="dosna_engine", help=hide,
                        default=None)
    parser.add_argument("--dosna_connection", dest="dosna_connection",
                        help=hide, default=None)
    parser.add_argument("--dosna_connection_options",
                        dest="dosna_connection_options", help=hide,
                        nargs='+', default=[])

    check_help = "Continue Savu processing from a checkpoint."
    choices = ['plugin', 'subplugin']
    parser.add_argument("--checkpoint", nargs="?", choices=choices,
                        const='plugin', help=check_help, default=None)

    args = parser.parse_args()
    __check_conditions(parser, args)
    return args


def __check_conditions(parser, args):
    if args.checkpoint and not args.folder:
        msg = "--checkpoint flag requires '-f folder_name', where folder_name"\
              " contains the partially completed Savu job.  The out_folder"\
              " should be the path to this folder."
        parser.error(msg)


def _set_options(args):
    """ Set run specific information in options dictionary.

    :params dict opt: input optional arguments (or defaults)
    :params args: input required arguments
    :returns options: optional and required arguments
    :rtype: dict
    """
    options = {}
    options['data_file'] = args.in_file
    options['process_file'] = args.process_list
    options['mode'] = args.mode
    options['template'] = args.template
    options['transport'] = 'basic' if args.mode == 'basic' else args.transport
    options['process_names'] = args.names
    options['verbose'] = args.verbose
    options['quiet'] = args.quiet
    options['cluster'] = args.cluster
    options['syslog_server'] = args.syslog
    options['syslog_port'] = args.syslog_port
    options['test_state'] = args.test_state
    options['lustre'] = args.lustre
    options['bllog'] = args.bllog
    options['email'] = args.email
    options['femail'] = args.femail
    options['system_params'] = args.system_params

    if args.folder:
        out_folder_name = os.path.basename(args.folder)
    else:
        out_folder_name = __create_folder_name(options['data_file'])

    out_folder_path = __create_output_folder(args.out_folder, out_folder_name)

    options['out_folder'] = out_folder_name
    options['out_path'] = out_folder_path

    basename = os.path.basename(args.in_file)
    options['datafile_name'] = os.path.splitext(basename)[0] if basename \
        else args.in_file.split(os.sep)[-2]

    inter_folder_path = __create_output_folder(args.tmp, out_folder_name)\
        if args.tmp else out_folder_path

    options['inter_path'] = inter_folder_path
    options['log_path'] = args.log if args.log else options['inter_path']
    options['nProcesses'] = len(options["process_names"].split(','))
    # DosNa related options
    options["dosna_backend"] = args.dosna_backend
    options["dosna_engine"] = args.dosna_engine
    options["dosna_connection"] = args.dosna_connection
    options["dosna_connection_options"] = args.dosna_connection_options

    options['checkpoint'] = args.checkpoint

    return options


def __create_folder_name(dpath):
    if os.path.isfile(dpath):
        dpath = os.path.splitext(dpath)[0]
    elif os.path.isdir(dpath):
        dpath = os.path.dirname(dpath)
    import time
    MPI.COMM_WORLD.barrier()
    timestamp = time.strftime("%Y%m%d%H%M%S")
    MPI.COMM_WORLD.barrier()
    return "_".join([timestamp, os.path.basename(dpath)])


def __create_output_folder(path, folder_name):
    folder = os.path.join(path, folder_name)
    if MPI.COMM_WORLD.rank == 0:
        if not os.path.exists(folder):
            os.makedirs(folder)
    return folder


def main(input_args=None):
    args = __option_parser()

    if input_args:
        args = input_args

    options = _set_options(args)
    pRunner = PluginRunner if options['mode'] == 'full' else BasicPluginRunner

    if options['nProcesses'] == 1:
        plugin_runner = pRunner(options)
        plugin_runner._run_plugin_list()
    else:
        try:
            plugin_runner = pRunner(options)
            plugin_runner._run_plugin_list()
        except Exception as error:
            print(error)
            traceback.print_exc(file=sys.stdout)
            MPI.COMM_WORLD.Abort(1)


if __name__ == '__main__':
    main()



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: multi_savu_loader
   :platform: Unix
   :synopsis: A class for loading multiple savu output nexus files.

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""

import h5py
import os
import tempfile
import copy

from savu.plugins.loaders.base_loader import BaseLoader
from savu.plugins.utils import register_plugin
from savu.plugins.loaders.savu_nexus_loader import SavuNexusLoader
from savu.data.data_structures.data_types.stitch_data import StitchData


@register_plugin
class MultiSavuLoader(BaseLoader):
    """
    A class to load multiple savu datasets in Nexus format into one dataset.

    :param file_name: The shared part of the name of each file\
        (not including .nxs). Default: None.
    :param data_path: Path to the data inside the \
        file. Default: 'entry1/tomo_entry/data/data'.
    :param stack_or_cat: Stack or concatenate the data\
        (4D and 3D respectively). Default: 'stack'.
    :param stack_or_cat_dim: Dimension to stack or concatenate. Default: 3.
    :param axis_label: New axis label, if required, in the form\
        'name.units'. Default: 'scan.number'.
    :param range: The start and end of file numbers. Default: [0, 10].
    :param name: Name associated with the data set. Default: 'tomo'.
    """

    def __init__(self, name='MultiSavuLoader'):
        super(MultiSavuLoader, self).__init__(name)

    def setup(self):
        savu = self._get_savu_loader()
        data_obj_list = self._get_data_objects(savu)
        data_obj = self.exp.create_data_object('in_data', 'savu')

        # dummy file
        filename = os.path.split('/')[-1] + '.h5'
        data_obj.backing_file = \
            h5py.File(tempfile.mkdtemp() + '/' + filename, 'a')

        stack_or_cat = self.parameters['stack_or_cat']
        dim = self.parameters['stack_or_cat_dim']
        data_obj.data = StitchData(data_obj_list, stack_or_cat, dim)

        self._set_axis_labels(data_obj, data_obj_list[0])
        self._set_patterns(data_obj, data_obj_list[0])

        data_obj.set_original_shape(data_obj.data.get_shape())
        return data_obj

    def _set_axis_labels(self, data_obj, savu_obj):
        axis_labels = savu_obj.get_axis_labels()
        if self.parameters['stack_or_cat'] == 'stack':
            name, unit = self.parameters['axis_label'].split('.')
            axis_labels.append({name: unit})
        data_obj.data_info.set('axis_labels', axis_labels)

    def _set_patterns(self, data_obj, savu_obj):
        patterns = savu_obj.data_info.get('data_patterns')

        if self.parameters['stack_or_cat'] == 'stack':
            for p in patterns:
                patterns[p]['slice_dims'] += (3,)
        data_obj.data_info.set('data_patterns', patterns)

    def _get_data_objects(self, savu):
        rrange = self.parameters['range']
        file_list = list(range(rrange[0], rrange[1]+1))
        file_path = copy.copy(self.exp.meta_data.get('data_file'))
        file_name = '' if self.parameters['file_name'] is None else\
            self.parameters['file_name']

        data_obj_list = []
        for i in file_list:
            this_file = file_path + file_name + str(i) + '.h5'
            self.exp.meta_data.set('data_file', this_file)
            savu.setup()
            data_obj_list.append(self.exp.index['in_data']['tomo'])
            self.exp.index['in_data'] = {}

        self.exp.meta_data.set('data_file', file_path)
        return data_obj_list

    def _get_savu_loader(self):
        savu = SavuLoader()
        savu.exp = self.exp
        savu._populate_default_parameters()
        savu.parameters['data_path'] = self.parameters['data_path']
        savu.parameters['name'] = self.parameters['name']
        return savu



# Copyright 2019 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
 .. module:: rescale_intensity to (0, 1) levels by stretching or shrinking the data intensity levels
   :platform: Unix
   :synopsis: performs stretching or shrinking the data intensity levels based on skimage rescale_intensity module

.. moduleauthor:: Daniil Kazantsev <scientificsoftware@diamond.ac.uk>
"""


from savu.plugins.plugin import Plugin
from savu.plugins.driver.cpu_plugin import CpuPlugin
from savu.plugins.utils import register_plugin

import skimage.exposure
import skimage.io
import numpy as np

@register_plugin
class DataRescale(Plugin, CpuPlugin):
    """
    The plugin performs stretching or shrinking the data intensity levels based on skimage rescale_intensity module. \
    Min-max scalars for rescaling can be passed from METADATA OR by providing as an input.
    
    :param min_value: the global minimum data value. Default: None.
    :param max_value: the global maximum data value. Default: None.
    """

    def __init__(self):
        super(DataRescale, self).__init__("DataRescale")

    def setup(self):
        in_dataset, out_dataset = self.get_datasets()
        out_dataset[0].create_dataset(in_dataset[0])
        in_pData, out_pData = self.get_plugin_datasets()
        in_pData[0].plugin_data_setup('VOLUME_XZ', 'single')
        out_pData[0].plugin_data_setup('VOLUME_XZ', 'single')

    def pre_process(self):
        data = self.get_in_datasets()[0]
        try:
            the_max = data.meta_data.get(['stats', 'max', 'pattern'])
        except KeyError:
            the_max = self.parameters['max_value']
        try:
            the_min = data.meta_data.get(['stats', 'min', 'pattern'])
        except KeyError:
            the_min = self.parameters['min_value']
        
        self._data_range = (the_min, the_max)

    def process_frames(self, data):
        if ((self._data_range[0] is not None) & (self._data_range[1] is not None)):
            # Rescale image to (0.0, 1.0) range
            resampled_image = skimage.exposure.rescale_intensity(data[0], in_range=self._data_range, out_range=(0.0, 1.0))
        else:
            resampled_image = data[0]
        return resampled_image
    
    def nInput_datasets(self):
        return 1
    
    def nOutput_datasets(self):
        return 1

    def get_max_frames(self):
        return 'single'



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: preview
   :platform: Unix
   :synopsis: This class deals with previewing (reduction) of the data and is\
   encapsulated in the Data class.

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>
"""
import numpy as np

from savu.core.utils import docstring_parameter
from savu.plugins.utils import parse_config_string as parse_str
import savu.data.data_structures.data_notes as notes


class Preview(object):
    """The Data class dynamically inherits from transport specific data class
    and holds the data array, along with associated information.
    """
    def __init__(self, data_obj):
        self._data_obj = data_obj
        self.revert_shape = None
        self.revert_axis_labels = None
        self.plugin_preview = None

    def get_data_obj(self):
        return self._data_obj

    @docstring_parameter(notes._set_preview_note.__doc__)
    def set_preview(self, preview_list, **kwargs):
        """ Reduces the data to be processed to a subset of the original.

        :param list preview: previewing parameters, where
            ``len(preview_list)`` equals the number of data dimensions.
        :keyword bool revert: revert input dataset to the original size after
         plugin processing.

        {0}
        """
        self.revert_shape = kwargs.get('revert', self.revert_shape)
        load = kwargs.get('load', False)
        shape = self.get_data_obj().get_shape()
        # for backward compatibility
        preview_list = parse_str(preview_list) if \
            isinstance(preview_list, str) else preview_list
        preview_list = self.__convert_nprocs(preview_list)

        if preview_list:
            preview_list = self._add_preview_defaults(preview_list)
            starts, stops, steps, chunks = \
                self._get_preview_indices(preview_list)
            shape_change = True
        else:
            starts, stops, steps, chunks = \
                [[0]*len(shape), shape, [1]*len(shape), [1]*len(shape)]
            shape_change = False

        self.__set_starts_stops_steps(starts, stops, steps, chunks,
                                      shapeChange=shape_change, load=load)
        self.__check_preview_indices()

    def __convert_nprocs(self, preview_list):
        for i in range(len(preview_list)):
            if preview_list[i] == 'nprocs':
                nprocs = self.get_data_obj().exp.meta_data.get('nProcesses')
                start = int(np.floor(nprocs / 2.0))
                end = int(np.ceil(nprocs / 2.0))
                preview_list[i] = f'mid-{start}:mid-{end}'
        return preview_list

    def _add_preview_defaults(self, plist):
        """ Fill in missing values in preview list entries.

        :param: preview list with entries of the form
            ``start[:stop:step:chunk]``
        :returns: preview list with missing values replaced by defaults
        :rtype: list
        """
        nEntries = 4
        #plist = [str(i) for i in plist] if isinstance(plist[0], int) else plist
        plist = [str(i) if isinstance(i, int) else i for i in plist]
        diff_len = [(nEntries - len(elem.split(':'))) for elem in plist]
        diff3 = [i for i in range(len(diff_len)) if diff_len[i] == 3]
        for dim in diff3:
            plist[dim] = plist[dim] + ':' + plist[dim] + '+1'
            diff_len[dim] = 2

        all_idx = [i for i in range(len(plist)) if plist[i] == ':']
        amend = [i for i in range(len(plist)) if diff_len and i not in all_idx]
        for idx in amend:
            plist[idx] += ':1'*diff_len[idx]
        return plist

    def _unset_preview(self):
        """ Unset preview (revert=True) if it was only required in the plugin.
        """
        self._data_obj.set_shape(self.revert_shape)
        self.set_preview([])
        self.revert_shape = None

    def _reset_preview(self):
        self.set_preview([])
        self.plugin_preview = None

    def __set_starts_stops_steps(self, starts, stops, steps, chunks,
                                 shapeChange=True, load=False):
        """ Add previewing params to data_info dictionary and set reduced
        shape.
        """

        set_mData = self.get_data_obj().data_info.set
        set_mData('starts', starts)
        set_mData('stops', stops)
        set_mData('steps', steps)
        set_mData('chunks', chunks)
        if shapeChange:
            self.__set_reduced_shape(starts, stops, steps, chunks)
            slice_list = self._get_preview_slice_list()
            self.get_data_obj().amend_axis_label_values(slice_list)
        if load:
            self.__add_preview_param('starts', starts)
            self.__add_preview_param('stops', stops)
            self.__add_preview_param('steps', steps)

    def __add_preview_param(self, name, value):
        entry = self.get_data_obj().get_name() + '_preview_' + name
        self.get_data_obj().exp.meta_data.set(entry, value)

    def _get_preview_indices(self, preview_list):
        """ Get preview_list ``starts``, ``stops``, ``steps``, ``chunks``
        separate components with integer values.

        :params: preview_list
        :returns: separate list of starts, stops, steps, chunks integer values
        :rtype: list(list(int))
        """
        starts = len(preview_list)*[None]
        stops = len(preview_list)*[None]
        steps = len(preview_list)*[None]
        chunks = len(preview_list)*[None]

        for i in range(len(preview_list)):
            if preview_list[i] == ':':
                preview_list[i] = '0:end:1:1'
            vals = preview_list[i].split(':')
            starts[i], stops[i], steps[i], chunks[i] = \
                self.convert_indices(vals, i)

        return starts, stops, steps, chunks

    def get_integer_entries(self, plist):
        """
        Convert Savu preview syntax to python slicing (similar) syntax, by
        replacing Savu Built-in constants.
        
        Parameters
        ----------
        plist : list
            A Savu data preview list to reduce the data dimensions.

        Returns
        -------
        list
            A Savu preview list containing integers and no strings.

        """
        if plist:
            vals = self._get_preview_indices(self._add_preview_defaults(plist))
        else:
            shape = self.get_data_obj().get_shape()
            vals = [[0]*len(shape), shape, [1]*len(shape), [1]*len(shape)]        
        return [':'.join(map(str, l)) for l in list(zip(*vals))]

    def convert_indices(self, idx, dim):
        """ convert keywords to integers.
        """
        dobj = self.get_data_obj()
        shape = dobj.get_shape()
        mid = np.clip(np.ceil(shape[dim] / 2.0).astype('int') - 1, 0, None)
        end = shape[dim]
        idx = [eval(equ, {"builtins": None}, {'mid': mid, 'end': end}) for equ in idx]
        idx = [idx[i] if idx[i] > -1 else shape[dim]+1+idx[i] for i in
               range(len(idx))]
        return idx

    def get_starts_stops_steps(self, key=None):
        """ Returns preview parameter ``starts``, ``stops``, ``steps``,
        ``chunks`` lists.

        :keyword str key: the list to return.
        :returns: if key is none return  separate preview_list components,
         where each list has length equal to number of dataset dimensions, else
         only the ``key`` list.
        :rtype: list(list(int))
        """
        mData = self.get_data_obj().data_info

        if 'starts' not in list(mData.get_dictionary().keys()):
            return None if key else [None]*4

        if key is not None:
            return mData.get(key)

        return [mData.get('starts'), mData.get('stops'), mData.get('steps'),
                mData.get('chunks')]

    def __check_preview_indices(self):
        starts, stops, steps, chunks = self.get_starts_stops_steps()
        nDims = len(starts)
        for i in range(nDims):
            if stops[i] <= starts[i]:
                raise Exception("Error in previewing parameters! "
                                "Check parameters that may alter data dimensions. "
                                "(axis={}, start={}, stop={})".format(
                                    i, starts[i], stops[i]))

    def __set_reduced_shape(self, starts, stops, steps, chunks):
        """ Set new shape if data is reduced by previewing.
        """
        dobj = self.get_data_obj()
        td = dobj._get_transport_data()
        orig_shape = dobj.get_shape()
        dobj.data_info.set('orig_shape', orig_shape)
        new_shape = []
        for dim in range(len(starts)):
            new_shape.append(np.prod((td._get_slice_dir_matrix(dim).shape)))
        dobj.set_shape(tuple(new_shape))

    def _get_preview_slice_list(self):
        """ Amend the axis label values based on the previewing parameters.
        """
        dobj = self.get_data_obj()
        td = dobj._get_transport_data()
        starts, stops, steps, chunks = self.get_starts_stops_steps()

        if starts is None:
            return None

        slice_list = []
        for dim in range(len(dobj.get_shape())):
            if chunks[dim] > 1:
                slice_list.append(
                    np.ravel(np.transpose(td._get_slice_dir_matrix(dim))))
            else:
                slice_list.append(slice(starts[dim], stops[dim], steps[dim]))
        return slice_list



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: hdf5_transport_data
   :platform: Unix
   :synopsis: A data transport class that is inherited by Data class at \
   runtime. It organises the slice list and moves the data.

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""

from savu.data.transport_data.slice_lists import \
    SliceLists, LocalData, GlobalData
from savu.data.transport_data.base_transport_data import BaseTransportData


class BasicTransportData(BaseTransportData, SliceLists):
    """
    The Hdf5TransportData class performs the organising and movement of data.
    """

    def __init__(self, data_obj, name='BasicTransportData'):
        super(BasicTransportData, self).__init__(data_obj)

    def _get_slice_lists_per_process(self, dtype):
        pData = self.data._get_plugin_data()
        pData._set_padding_dict()
        self.pad = True if pData.padding else False
        local_dict = LocalData(dtype, self)._get_dict()
        if dtype == 'in':
            gdict = GlobalData(dtype, self)._get_dict()
            local_dict['frames'] = gdict['frames']
            local_dict['current'] = gdict['current']
        return local_dict

    def _calc_max_frames_transfer(self, nFrames):
        t1, t2 = self._calc_max_frames_transfer_single(nFrames)
        return t1, t2

    def _get_padded_data(self, input_slice_list):
        data = self.data.data[input_slice_list]
        return data



# Copyright 2019 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: A plugin to calculate global statistics (max, min sum, mean) of the input data
   :platform: Unix
   :synopsis: A plugin to calculate global statistics (max, min, sum, mean) of the input data

.. moduleauthor:: Daniil Kazantsev <scientificsoftware@diamond.ac.uk>
"""

from savu.plugins.plugin import Plugin
from savu.plugins.driver.cpu_plugin import CpuPlugin
from savu.plugins.utils import register_plugin

import numpy as np

@register_plugin
class GetDataStatistics(Plugin, CpuPlugin):
    """
    Collect input data global statistcs.

    :param out_datasets: The default names . Default: ['data, data_statistics'].
    """

    def __init__(self):
        super(GetDataStatistics, self).__init__("GetDataStatistics")

    def setup(self):

        in_dataset, out_dataset = self.get_datasets()
        in_pData, out_pData = self.get_plugin_datasets()
        in_pData[0].plugin_data_setup('VOLUME_XZ', 'single')

        out_dataset[0].create_dataset(in_dataset[0])
        out_pData[0].plugin_data_setup('VOLUME_XZ', 'single')
        
        fullData = in_dataset[0]
        slice_dirs = list(in_dataset[0].get_slice_dimensions())
        self.new_shape = (np.prod(np.array(fullData.get_shape())[slice_dirs]), 4)
        out_dataset[1].create_dataset(shape=self.new_shape,
                                      axis_labels=['stattype', 'value'],
                                      remove=True,
                                      transport='hdf5')
        out_dataset[1].add_pattern("METADATA", core_dims=(1,), slice_dims=(0,))
        out_pData[1].plugin_data_setup('METADATA', self.get_max_frames())

    def process_frames(self, data):
        data_temp = data[0]
        indices = np.where(np.isnan(data_temp))
        data_temp[indices] = 0.0
        # collecting values for each slice
        slice_values = [np.max(data_temp), np.min(data_temp), np.sum(data_temp), np.mean(data_temp)]
        return [data_temp, np.array([slice_values])]

    def post_process(self):
        data, slice_values = self.get_out_datasets()
        all_slice_values = slice_values.data[...]
        max_stat = np.max(all_slice_values[:,0])
        min_stat = np.max(all_slice_values[:,1])
        sum_stat = np.sum(all_slice_values[:,2])
        mean_stat = np.sum(all_slice_values[:,3])
        data.meta_data.set(['stats', 'max', 'pattern'], max_stat)
        data.meta_data.set(['stats', 'min', 'pattern'], min_stat)
        data.meta_data.set(['stats', 'sum', 'pattern'], sum_stat)
        data.meta_data.set(['stats', 'mean', 'pattern'], mean_stat)

    def nInput_datasets(self):
        return 1
    def nOutput_datasets(self):
        return 2
    def get_max_frames(self):
        return 'single'



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: checkpoint_tests
   :platform: Unix
   :synopsis: Checking Savu checkpointing works correctly

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""

import unittest
from savu.test import test_utils as tu
from savu.test.base_checkpoint_test import BaseCheckpointTest


class CheckpointTest(BaseCheckpointTest, unittest.TestCase):

    def setUp(self):
        super(CheckpointTest, self).setUp()

    def tearDown(self):
        super(CheckpointTest, self).tearDown()

    def set_plist(self, exp):
        plist = exp.meta_data.plugin_list.plugin_list
        plist = [p['name'] for p in plist[1:]]
        plist[-1] = 'final_result_tomo'
        return plist

    def set_data_options(self):
        options = tu.set_options(tu.get_test_data_path('24737.nxs'))
        options['process_file'] = \
            tu.get_test_process_path('checkpoint_process2.nxs')
        return options

if __name__ == "__main__":
    unittest.main()



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: base_astra_recon
   :platform: Unix
   :synopsis: A base for all Astra toolbox reconstruction algorithms
.. moduleauthor:: Mark Basham <scientificsoftware@diamond.ac.uk>
"""

import astra
import numpy as np

from savu.plugins.reconstructions.base_recon import BaseRecon
from savu.data.plugin_list import CitationInformation


class BaseAstraRecon(BaseRecon):
    """
    A Plugin to perform Astra toolbox reconstruction

    :u*param n_iterations: Number of Iterations - only valid for iterative \
        algorithms. Default: 1.
    """

    def __init__(self, name='BaseAstraRecon'):
        super(BaseAstraRecon, self).__init__(name)
        self.res = False

    def setup(self):
        self.alg = self.parameters['algorithm']
        self.get_max_frames = self._get_multiple if '3D' in self.alg else self._get_single

        super(BaseAstraRecon, self).setup()
        out_dataset = self.get_out_datasets()

        # if res_norm is required then setup another output dataset
        if len(out_dataset) == 2:
            self.res = True
            out_pData = self.get_plugin_out_datasets()
            in_data = self.get_in_datasets()[0]
            dim_detX = \
                in_data.get_data_dimension_by_axis_label('y', contains=True)

            nIts = self.parameters['n_iterations']
            nIts = nIts if isinstance(nIts, list) else [nIts]
            self.len_res = max(nIts)
            shape = (in_data.get_shape()[dim_detX], max(nIts))

            label = ['vol_y.voxel', 'iteration.number']
            pattern = {'name': 'SINOGRAM', 'slice_dims': (0,),
                       'core_dims': (1,)}

            out_dataset[1].create_dataset(axis_labels=label, shape=shape)
            out_dataset[1].add_pattern(pattern['name'],
                                       slice_dims=pattern['slice_dims'],
                                       core_dims=pattern['core_dims'])
            out_pData[1].plugin_data_setup(
                pattern['name'], self.get_max_frames())

    def pre_process(self):
        self.alg = self.parameters['algorithm']
        self.iters = self.parameters['n_iterations']

        if '3D' in self.alg:
            self.setup_3D()
            self.process_frames = self.astra_3D_recon
        else:
            self.setup_2D()
            self.process_frames = self.astra_2D_recon

    def setup_2D(self):
        pData = self.get_plugin_in_datasets()[0]
        self.dim_detX = \
            pData.get_data_dimension_by_axis_label('x', contains=True)
        self.dim_rot = \
            pData.get_data_dimension_by_axis_label('rot', contains=True)

        self.sino_shape = pData.get_shape()
        self.nDims = len(self.sino_shape)
        self.nCols = self.sino_shape[self.dim_detX]
        self.set_mask(self.sino_shape)

    def set_mask(self, shape):
        l = self.get_plugin_out_datasets()[0].get_shape()[0]
        c = np.linspace(-l / 2.0, l / 2.0, l)
        x, y = np.meshgrid(c, c)

        ratio = self.parameters['ratio']
        if isinstance(ratio, list) or isinstance(ratio, tuple):
            ratio_mask = ratio[0]
            outer_mask = ratio[1]
            if isinstance(outer_mask, str):
                outer_mask = np.nan
        else:
            ratio_mask = ratio
            outer_mask = np.nan
        r = (l - 1) * ratio_mask
        outer_pad = True if self.parameters['outer_pad'] and self.padding_alg\
            else False
        if not outer_pad:
            self.manual_mask = \
                np.array((x**2 + y**2 < (r / 2.0)**2), dtype=np.float)
            self.manual_mask[self.manual_mask == 0] = outer_mask
        else:
            self.manual_mask = False

    def astra_2D_recon(self, data):
        sino = data[0]
        cor, angles, vol_shape, init = self.get_frame_params()
        angles = np.deg2rad(angles)
        if self.res:
            res = np.zeros(self.len_res)
        # create volume geom
        vol_geom = astra.create_vol_geom(vol_shape)
        # create projection geom
        det_width = sino.shape[self.dim_detX]
        proj_geom = astra.create_proj_geom('parallel', 1.0, det_width, angles)
        sino = np.transpose(sino, (self.dim_rot, self.dim_detX))

        # create sinogram id
        sino_id = astra.data2d.create("-sino", proj_geom, sino)
        # create reconstruction id
        if init is not None:
            rec_id = astra.data2d.create('-vol', vol_geom, init)
        else:
            rec_id = astra.data2d.create('-vol', vol_geom)

#        if self.mask_id:
#            self.mask_id = astra.data2d.create('-vol', vol_geom, self.mask)
        # setup configuration options
        cfg = self.set_config(rec_id, sino_id, proj_geom, vol_geom)
        # create algorithm id
        alg_id = astra.algorithm.create(cfg)
        # run algorithm
        if self.res:
            for j in range(self.iters):
                # Run a single iteration
                astra.algorithm.run(alg_id, 1)
                res[j] = astra.algorithm.get_res_norm(alg_id)
        else:
            astra.algorithm.run(alg_id, self.iters)
        # get reconstruction matrix

        if self.manual_mask is not False:
            recon = self.manual_mask * astra.data2d.get(rec_id)
        else:
            recon = astra.data2d.get(rec_id)

        # delete geometry
        self.delete(alg_id, sino_id, rec_id, False)
        return [recon, res] if self.res else recon

    def set_config(self, rec_id, sino_id, proj_geom, vol_geom):
        cfg = astra.astra_dict(self.alg)
        cfg['ReconstructionDataId'] = rec_id
        cfg['ProjectionDataId'] = sino_id
        if 'FBP' in self.alg:
            fbp_filter = self.parameters['FBP_filter'] if 'FBP_filter' in \
                list(self.parameters.keys()) else 'none'
            cfg['FilterType'] = fbp_filter
        if 'projector' in list(self.parameters.keys()):
            proj_id = astra.create_projector(
                self.parameters['projector'], proj_geom, vol_geom)
            cfg['ProjectorId'] = proj_id
        cfg = self.set_options(cfg)
        return cfg

    def delete(self, alg_id, sino_id, rec_id, proj_id):
        astra.algorithm.delete(alg_id)
        astra.data2d.delete(sino_id)
        astra.data2d.delete(rec_id)
        if proj_id:
            astra.projector.delete(proj_id)

    def get_padding_algorithms(self):
        """ A list of algorithms that allow the data to be padded. """
        return ['FBP', 'FBP_CUDA']

    def _get_single(self):
        return 'single'

    def _get_multiple(self):
        return 'multiple'

    def get_citation_information(self):
        cite_info1 = CitationInformation()
        cite_info1.name = 'citation1'
        cite_info1.description = \
            ("The tomography reconstruction algorithm used in this processing \
             pipeline is part of the ASTRA Toolbox")
        cite_info1.bibtex = \
            ("@article{van2016fast,\n" +
             "title={Fast and flexible X-ray tomography using the ASTRA \
             toolbox},\n" +
             "author={van Aarle, Wim and Palenstijn, Willem Jan and Cant, \
             Jeroen and Janssens, Eline and Bleichrodt, Folkert and \
             Dabravolski, Andrei and De Beenhouwer, Jan and Batenburg, K Joost\
             and Sijbers, Jan},\n" +
             "journal={Optics Express},\n" +
             "volume={24},\n" +
             "number={22},\n" +
             "pages={25129--25147},\n" +
             "year={2016},\n" +
             "publisher={Optical Society of America}\n" +
             "}")
        cite_info1.endnote = \
            ("%0 Journal Article\n" +
             "%T Fast and flexible X-ray tomography using the ASTRA \
             toolbox\n" +
             "%A van Aarle, Wim\n" +
             "%A Palenstijn, Willem Jan\n" +
             "%A Cant, Jeroen\n" +
             "%A Janssens, Eline\n" +
             "%A Bleichrodt, Folkert\n" +
             "%A Dabravolski, Andrei\n" +
             "%A De Beenhouwer, Jan\n" +
             "%A Batenburg, K Joost\n" +
             "%A Sijbers, Jan\n" +
             "%J Optics Express\n" +
             "%V 24\n" +
             "%N 22\n" +
             "%P 25129-25147\n" +
             "%@ 1094-4087\n" +
             "%D 2016\n" +
             "%I Optical Society of America\n")
        cite_info1.doi = "doi: 10.1364/OE.24.025129"

        cite_info2 = CitationInformation()
        cite_info2.name = 'citation2'
        cite_info2.description = \
            ("The tomography reconstruction algorithm used in this processing \
             pipeline is part of the ASTRA Toolbox")
        cite_info2.bibtex = \
            ("@article{van2015astra,\n" +
             "title={The ASTRA Toolbox: A platform for advanced algorithm \
             development in electron tomography},\n" +
             "author={van Aarle, Wim and Palenstijn, Willem Jan and \
             De Beenhouwer, Jan and Altantzis, Thomas and Bals, Sara and \
             Batenburg, K Joost and Sijbers, Jan},\n" +
             "journal={Ultramicroscopy},\n" +
             "volume={157},\n" +
             "pages={35--47},\n" +
             "year={2015},\n" +
             "publisher={Elsevier}\n" +
             "}")
        cite_info2.endnote = \
            ("%0 Journal Article\n" +
             "%T Numerical removal of ring artifacts in microtomography\n" +
             "%A Raven, Carsten\n" +
             "%J Review of scientific instruments\n" +
             "%V 69\n" +
             "%N 8\n" +
             "%P 2978-2980\n" +
             "%@ 0034-6748\n" +
             "%D 1998\n" +
             "%I AIP Publishing")
        cite_info2.doi = "doi: 10.1364/OE.24.025129"

        return [cite_info1, cite_info2]



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: nxtomo_loader
   :platform: Unix
   :synopsis: A class for loading standard tomography data in Nexus format.

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""

import h5py
import logging
import numpy as np

import savu.core.utils as cu
from savu.plugins.loaders.base_loader import BaseLoader
from savu.plugins.utils import register_plugin
from savu.data.data_structures.data_types.data_plus_darks_and_flats \
    import ImageKey, NoImageKey

@register_plugin
class NxtomoLoader(BaseLoader):
    """
    A class to load tomography data from a hdf5 file

    :param name: The name assigned to the dataset. Default: 'tomo'.
    :param data_path: Path to the data inside the \
        file. Default: 'entry1/tomo_entry/data/data'.
    :param image_key_path: Path to the image key entry inside the nxs \
        file. Set this parameter to "None" if use this loader for radiography\
        . Default: 'entry1/tomo_entry/instrument/detector/image_key'.
    :param dark: Optional path to the dark field data file, nxs path and \
        scale value. Default: [None, None, 1].
    :param flat: Optional Path to the flat field data file, nxs path and \
        scale value. Default: [None, None, 1].
    :param angles: A python statement to be evaluated or a file. Default: None.
    :param 3d_to_4d: If this is 4D data stored in 3D then pass an integer \
        value equivalent to the number of projections per 180-degree scan\
        . Default: False.
    :param ignore_flats: List of batch numbers of flats (start at 1) to \
        ignore. Default: None.
    """

    def __init__(self, name='NxtomoLoader'):
        super(NxtomoLoader, self).__init__(name)
        self.warnings = []

    def log_warning(self, msg):
        logging.warning(msg)
        self.warnings.append(msg)

    def setup(self):
        exp = self.get_experiment()
        data_obj = exp.create_data_object('in_data', self.parameters['name'])

        data_obj.backing_file = self._get_data_file()

        data_obj.data = self._get_h5_entry(
                data_obj.backing_file, self.parameters['data_path'])

        self._set_dark_and_flat(data_obj)

        self.nFrames = self.__get_nFrames(data_obj)
        if self.nFrames > 1:
            self.__setup_4d(data_obj)
            self.__setup_3d_to_4d(data_obj, self.nFrames)
        else:
            if len(data_obj.data.shape) == 3:
                self._setup_3d(data_obj)
            else:
                self.__setup_4d(data_obj)
            data_obj.set_original_shape(data_obj.data.shape)
        self._set_rotation_angles(data_obj)

        try:
            control = self._get_h5_path(
                    data_obj.backing_file, 'entry1/tomo_entry/control/data')
            data_obj.meta_data.set("control", control[...])
        except Exception:
            self.log_warning("No Control information available")

        nAngles = len(data_obj.meta_data.get('rotation_angle'))
        self.__check_angles(data_obj, nAngles)

        self.set_data_reduction_params(data_obj)
        data_obj.data._set_dark_and_flat()

    def _get_h5_entry(self, filename, path):
        if path in filename:
            return filename[path]
        else:
            raise Exception("Path %s not found in %s" % (path, filename))

    def __get_nFrames(self, dObj):
        if self.parameters['3d_to_4d'] is False:
            return 0
        if self.parameters['3d_to_4d'] is True:
            try:
                # for backwards compatibility
                n_frames = eval(self.parameters["angles"], {"builtins": None, "np": np})
                return np.array(n_frames).shape[0]
            except Exception:
                raise Exception("Please specify the angles, or the number of "
                                "frames per scan (via 3d_to_4d param) in the loader.")
        if isinstance(self.parameters['3d_to_4d'], int):
            return self.parameters['3d_to_4d']
        else:
            raise Exception("Unknown value for loader parameter '3d_to_4d', "
                            "please specify an integer value")

    def _setup_3d(self, data_obj):
        logging.debug("Setting up 3d tomography data.")
        rot = 0
        detY = 1
        detX = 2
        data_obj.set_axis_labels('rotation_angle.degrees',
                                 'detector_y.pixel',
                                 'detector_x.pixel')

        data_obj.add_pattern('PROJECTION', core_dims=(detX, detY),
                             slice_dims=(rot,))
        data_obj.add_pattern('SINOGRAM', core_dims=(detX, rot),
                             slice_dims=(detY,))
        data_obj.add_pattern('TANGENTOGRAM', core_dims=(rot, detY),
                             slice_dims=(detX,))

    def __setup_3d_to_4d(self, data_obj, n_frames):
        logging.debug("setting up 4d tomography data from 3d input.")
        from savu.data.data_structures.data_types.map_3dto4d_h5 \
            import Map3dto4dh5

        all_angles = data_obj.data.shape[0]
        if all_angles % n_frames != 0:
            self.log_warning("There are not a complete set of scans in this file, "
                             "loading complete ones only")
        data_obj.data = Map3dto4dh5(data_obj, n_frames)
        data_obj.set_original_shape(data_obj.data.get_shape())

    def __setup_4d(self, data_obj):
        logging.debug("setting up 4d tomography data.")
        rot = 0
        detY = 1
        detX = 2
        scan = 3

        data_obj.set_axis_labels('rotation_angle.degrees', 'detector_y.pixel',
                                 'detector_x.pixel', 'scan.number')

        data_obj.add_pattern('PROJECTION', core_dims=(detX, detY),
                             slice_dims=(rot, scan))
        data_obj.add_pattern('SINOGRAM', core_dims=(detX, rot),
                             slice_dims=(detY, scan))
        data_obj.add_pattern('TANGENTOGRAM', core_dims=(rot, detY),
                             slice_dims=(detX, scan))
        data_obj.add_pattern('SINOMOVIE', core_dims=(detX, rot, scan),
                             slice_dims=(detY,))

    def _set_dark_and_flat(self, data_obj):
        flat = self.parameters['flat'][0]
        dark = self.parameters['dark'][0]

        if not flat or not dark:
            self.__find_dark_and_flat(data_obj, flat=flat, dark=dark)
        else:
            self.__set_separate_dark_and_flat(data_obj)

    def __find_dark_and_flat(self, data_obj, flat=None, dark=None):
        ignore = self.parameters['ignore_flats'] if \
            self.parameters['ignore_flats'] else None
        try:
            image_key = data_obj.backing_file[
                self.parameters['image_key_path']][...]
            data_obj.data = \
                ImageKey(data_obj, image_key, 0, ignore=ignore)
        except KeyError:
            self.log_warning("An image key was not found.")
            try:
                data_obj.data = NoImageKey(data_obj, None, 0)
                entry = 'entry1/tomo_entry/instrument/detector/'
                data_obj.data._set_flat_path(entry + 'flatfield')
                data_obj.data._set_dark_path(entry + 'darkfield')
            except KeyError:
                self.log_warning("Dark/flat data was not found in input file.")
        data_obj.data._set_dark_and_flat()
        if dark:
            data_obj.data.update_dark(dark)
        if flat:
            data_obj.data.update_flat(flat)

    def __set_separate_dark_and_flat(self, data_obj):
        try:
            image_key = data_obj.backing_file[
                'entry1/tomo_entry/instrument/detector/image_key'][...]
        except:
            image_key = None
        data_obj.data = NoImageKey(data_obj, image_key, 0)
        self.__set_data(data_obj, 'flat', data_obj.data._set_flat_path)
        self.__set_data(data_obj, 'dark', data_obj.data._set_dark_path)

    def __set_data(self, data_obj, name, func):
        path, entry, scale = self.parameters[name]

        if path.split('/')[0] == 'test_data':
            import os
            path = \
                os.path.dirname(os.path.abspath(__file__))+'/../../../../'+path

        ffile = h5py.File(path, 'r')
        try:
            image_key = \
                ffile['entry1/tomo_entry/instrument/detector/image_key'][...]
            func(ffile[entry], imagekey=image_key)
        except:
            func(ffile[entry])

        data_obj.data._set_scale(name, scale)

    def _set_rotation_angles(self, data_obj):
        angles = self.parameters['angles']
        if angles is None:
            angles = 'entry1/tomo_entry/data/rotation_angle'

        nxs_angles = self.__get_angles_from_nxs_file(data_obj, angles)

        if nxs_angles is None:
            try:
                angles = eval(angles)
            except Exception as e:
                logging.warning(e.message)
                try:
                    angles = np.loadtxt(angles)
                except Exception as e:
                    logging.debug(e.message)
                    self.log_warning("No angles found so evenly distributing them "
                                     "between 0 and 180 degrees")
                    angles = np.linspace(0, 180, data_obj.get_shape()[0])
        else:
            angles = nxs_angles
        data_obj.meta_data.set("rotation_angle", angles)
        return len(angles)

    def __get_angles_from_nxs_file(self, data_obj, path):
        if path in data_obj.backing_file:
            idx = data_obj.data.get_image_key() == 0 if \
                isinstance(data_obj.data, ImageKey) else slice(None)
            return data_obj.backing_file[path][idx]
        else:
            self.log_warning("No rotation angle entry found in input file.")
            return None

    def _get_data_file(self):
        data = self.exp.meta_data.get("data_file")
        return h5py.File(data, 'r')

    def __check_angles(self, data_obj, n_angles):
        data_angles = data_obj.data.get_shape()[0]
        if data_angles != n_angles:
            # FIXME problem with this
            if self.nFrames > 1:
                rot_angles = data_obj.meta_data.get("rotation_angle")
                try:
                    full_rotations = n_angles/data_angles
                    cleaned_size = full_rotations*data_angles
                    if cleaned_size != n_angles:
                        rot_angles = rot_angles[0:cleaned_size]
                        self.log_warning("the angle list has more values than expected in it")
                    rot_angles = np.reshape(
                            rot_angles, [full_rotations, data_angles])
                    data_obj.meta_data.set("rotation_angle",
                                           np.transpose(rot_angles))
                    return
                except:
                    pass
            raise Exception("The number of angles %s does not match the data "
                            "dimension length %s" % (n_angles, data_angles))

    def executive_summary(self):
        """ Provide a summary to the user for the result of the plugin.

        e.g.
         - Warning, the sample may have shifted during data collection
         - Filter operated normally

        :returns:  A list of string summaries
        """
        if len(self.warnings) == 0:
            return ["Nothing to Report"]
        else:
            return self.warnings



# Copyright 2015 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: base_transport
   :platform: Unix
   :synopsis: A BaseTransport class which implements functions that control\
   the interaction between the data and plugin layers.

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""

import os
import time
import copy
import h5py
import logging
import numpy as np

import savu.core.utils as cu
import savu.plugins.utils as pu
from savu.data.data_structures.data_types.base_type import BaseType

NX_CLASS = 'NX_class'


class BaseTransport(object):
    """
    Implements functions that control the interaction between the data and
    plugin layers.
    """

    def __init__(self):
        self.pDict = None
        self.no_processing = False

    def _transport_initialise(self, options):
        """
        Any initial setup required by the transport mechanism on start up.\
        This is called before the experiment is initialised.
        """
        raise NotImplementedError("transport_control_setup needs to be "
                                  "implemented in %s", self.__class__)

    def _transport_update_plugin_list(self):
        """
        This method provides an opportunity to add or remove items from the
        plugin list before plugin list check.
        """
        pass

    def _transport_pre_plugin_list_run(self):
        """
        This method is called after all datasets have been created but BEFORE
        the plugin list is processed.
        """
        pass

    def _transport_load_plugin(self, exp, plugin_dict):
        """ This method is called before each plugin is loaded """
        return pu.plugin_loader(exp, plugin_dict)

    def _transport_pre_plugin(self):
        """
        This method is called directly BEFORE each plugin is executed, but \
        after the plugin is loaded.
        """
        pass

    def _transport_post_plugin(self):
        """
        This method is called directly AFTER each plugin is executed.
        """
        pass

    def _transport_post_plugin_list_run(self):
        """
        This method is called AFTER the full plugin list has been processed.
        """
        pass

    def _transport_terminate_dataset(self, data):
        """ A dataset that will subequently be removed by the framework.

        :param Data data: A data object to finalise.
        """
        pass

    def process_setup(self, plugin):
        pDict = {}
        pDict['in_data'], pDict['out_data'] = plugin.get_datasets()
        pDict['in_sl'] = self._get_all_slice_lists(pDict['in_data'], 'in')
        pDict['out_sl'] = self._get_all_slice_lists(pDict['out_data'], 'out')
        pDict['nIn'] = list(range(len(pDict['in_data'])))
        pDict['nOut'] = list(range(len(pDict['out_data'])))
        pDict['nProc'] = len(pDict['in_sl']['process'])
        if 'transfer' in list(pDict['in_sl'].keys()):
            pDict['nTrans'] = len(pDict['in_sl']['transfer'][0])
        else:
            pDict['nTrans'] = 1
        pDict['squeeze'] = self._set_functions(pDict['in_data'], 'squeeze')
        pDict['expand'] = self._set_functions(pDict['out_data'], 'expand')

        frames = [f for f in pDict['in_sl']['frames']]
        self._set_global_frame_index(plugin, frames, pDict['nProc'])
        self.pDict = pDict

    def _transport_process(self, plugin):
        """ Organise required data and execute the main plugin processing.

        :param plugin plugin: The current plugin instance.
        """
        logging.info("transport_process initialise")
        pDict, result, nTrans = self._initialise(plugin)
        logging.info("transport_process get_checkpoint_params")
        cp, sProc, sTrans = self.__get_checkpoint_params(plugin)

        count = 0  # temporary solution
        prange = list(range(sProc, pDict['nProc']))
        kill = False
        for count in range(sTrans, nTrans):
            end = True if count == nTrans-1 else False
            self._log_completion_status(count, nTrans, plugin.name)

            # get the transfer data
            logging.info("Transferring the data")
            transfer_data = self._transfer_all_data(count)

            # loop over the process data
            logging.info("process frames loop")
            result, kill = self._process_loop(
                    plugin, prange, transfer_data, count, pDict, result, cp)

            logging.info("Returning the data")
            self._return_all_data(count, result, end)

            if kill:
                return 1

        if not kill:
            cu.user_message("%s - 100%% complete" % (plugin.name))

    def _process_loop(self, plugin, prange, tdata, count, pDict, result, cp):
        kill_signal = False
        for i in prange:
            if cp and cp.is_time_to_checkpoint(self, count, i):
                # kill signal sent so stop the processing
                return result, True
            data = self._get_input_data(plugin, tdata, i, count)
            res = self._get_output_data(
                    plugin.plugin_process_frames(data), i)

            for j in pDict['nOut']:
                if res is not None:
                    out_sl = pDict['out_sl']['process'][i][j]
                    result[j][out_sl] = res[j]
                else:
                    result[j] = None
        return result, kill_signal

    def __get_checkpoint_params(self, plugin):
        cp = self.exp.checkpoint
        if cp:
            cp._initialise(plugin.get_communicator())
            return cp, cp.get_proc_idx(), cp.get_trans_idx()
        return None, 0, 0

    def _initialise(self, plugin):
        self.process_setup(plugin)
        pDict = self.pDict
        result = [np.empty(d._get_plugin_data().get_shape_transfer(),
                           dtype=np.float32) for d in pDict['out_data']]
        # loop over the transfer data
        nTrans = pDict['nTrans']
        self.no_processing = True if not nTrans else False
        return pDict, result, nTrans

    def _log_completion_status(self, count, nTrans, name):
        percent_complete: float = count / (nTrans * 0.01)
        cu.user_message("%s - %3i%% complete" % (name, percent_complete))

    def _transport_checkpoint(self):
        """ The framework has determined it is time to checkpoint.  What
        should the transport mechanism do? Override if appropriate. """
        return False

    def _transport_kill_signal(self):
        """ 
        An opportunity to send a kill signal to the framework.  Return
        True or False. """
        return False

    def _get_all_slice_lists(self, data_list, dtype):
        """ 
        Get all slice lists for the current process.

        :param list(Data) data_list: Datasets
        :returns: A list of dictionaries containing slice lists for each \
            dataset
        :rtype: list(dict)
        """
        sl_dict = {}
        for data in data_list:
            sl = data._get_transport_data()._get_slice_lists_per_process(dtype)
            for key, value in sl.items():
                if key not in sl_dict:
                    sl_dict[key] = [value]
                else:
                    sl_dict[key].append(value)

        for key in [k for k in ['process', 'unpad'] if k in list(sl_dict.keys())]:
            nData = list(range(len(sl_dict[key])))
            #rep = range(len(sl_dict[key][0]))
            sl_dict[key] = [[sl_dict[key][i][j] for i in nData if j < len(sl_dict[key][i])] for j in range(len(sl_dict[key][0]))]
        return sl_dict

    def _transfer_all_data(self, count):
        """ 
        Transfer data from file and pad if required.

        :param int count: The current frame index.
        :returns: All data for this frame and associated padded slice lists
        :rtype: list(np.ndarray), list(tuple(slice))
        """
        pDict = self.pDict
        data_list = pDict['in_data']

        if 'transfer' in list(pDict['in_sl'].keys()):
            slice_list = \
                [pDict['in_sl']['transfer'][i][count] for i in pDict['nIn']]
        else:
            slice_list = [slice(None)]*len(pDict['nIn'])

        section = []
        for idx in range(len(data_list)):
            section.append(data_list[idx]._get_transport_data().
                           _get_padded_data(slice_list[idx]))
        return section

    def _get_input_data(self, plugin, trans_data, nproc, ntrans):
        data = []
        current_sl = []
        for d in self.pDict['nIn']:
            in_sl = self.pDict['in_sl']['process'][nproc][d]
            data.append(self.pDict['squeeze'][d](trans_data[d][in_sl]))
            entry = ntrans*self.pDict['nProc'] + nproc
            if entry < len(self.pDict['in_sl']['current'][d]):
                current_sl.append(self.pDict['in_sl']['current'][d][entry])
            else:
                current_sl.append(self.pDict['in_sl']['current'][d][-1])
        plugin.set_current_slice_list(current_sl)
        return data

    def _get_output_data(self, result, count):
        if result is None:
            return
        unpad_sl = self.pDict['out_sl']['unpad'][count]
        result = result if isinstance(result, list) else [result]
        for j in self.pDict['nOut']:
            result[j] = self.pDict['expand'][j](result[j])[unpad_sl[j]]
        return result

    def _return_all_data(self, count, result, end):
        """ 
        Transfer plugin results for current frame to backing files.

        :param int count: The current frame index.
        :param list(np.ndarray) result: plugin results
        :param bool end: True if this is the last entry in the slice list.
        """
        pDict = self.pDict
        data_list = pDict['out_data']

        slice_list = None
        if 'transfer' in list(pDict['out_sl'].keys()):
            slice_list = \
                [pDict['out_sl']['transfer'][i][count] for i in pDict['nOut'] \
                     if len(pDict['out_sl']['transfer'][i]) > count]

        result = [result] if type(result) is not list else result

        for idx in range(len(data_list)):
            if result[idx] is not None:
                if slice_list:
                    temp = self._remove_excess_data(
                            data_list[idx], result[idx], slice_list[idx])
                    data_list[idx].data[slice_list[idx]] = temp
                else:
                    data_list[idx].data = result[idx]

    def _set_global_frame_index(self, plugin, frame_list, nProc):
        """ Convert the transfer global frame index to a process global frame
            index.
        """
        process_frames = []
        for f in frame_list:
            if len(f):
                process_frames.append(list(range(f[0]*nProc, (f[-1]+1)*nProc)))

        process_frames = np.array(process_frames)
        nframes = plugin.get_plugin_in_datasets()[0].get_total_frames()
        process_frames[process_frames >= nframes] = nframes - 1
        frames = process_frames[0] if process_frames.size else process_frames
        plugin.set_global_frame_index(frames)

    def _set_functions(self, data_list, name):
        """ Create a dictionary of functions to remove (squeeze) or re-add
        (expand) dimensions, of length 1, from each dataset in a list.

        :param list(Data) data_list: Datasets
        :param str name: 'squeeze' or 'expand'
        :returns: A dictionary of lambda functions
        :rtype: dict
        """
        str_name = 'self.' + name + '_output'
        function = {'expand': self.__create_expand_function,
                    'squeeze': self.__create_squeeze_function}
        ddict = {}
        for i in range(len(data_list)):
            ddict[i] = {i: str_name + str(i)}
            ddict[i] = function[name](data_list[i])
        return ddict

    def __create_expand_function(self, data):
        """ Create a function that re-adds missing dimensions of length 1.

        :param Data data: Dataset
        :returns: expansion function
        :rtype: lambda
        """
        slice_dirs = data.get_slice_dimensions()
        n_core_dirs = len(data.get_core_dimensions())
        new_slice = [slice(None)]*len(data.get_shape())
        possible_slices = [copy.copy(new_slice)]

        pData = data._get_plugin_data()
        if pData._get_rank_inc():
            possible_slices[0] += [0]*pData._get_rank_inc()

        if len(slice_dirs) > 1:
            for sl in slice_dirs[1:]:
                new_slice[sl] = None
        possible_slices.append(copy.copy(new_slice))
        new_slice[slice_dirs[0]] = None
        possible_slices.append(copy.copy(new_slice))
        possible_slices = possible_slices[::-1]
        return lambda x: x[tuple(possible_slices[len(x.shape)-n_core_dirs])]

    def __create_squeeze_function(self, data):
        """ Create a function that removes dimensions of length 1.

        :param Data data: Dataset
        :returns: squeeze function
        :rtype: lambda
        """
        pData = data._get_plugin_data()
        max_frames = pData._get_max_frames_process()

        pad = True if pData.padding and data.get_slice_dimensions()[0] in \
            list(pData.padding._get_padding_directions().keys()) else False

        n_core_dims = len(data.get_core_dimensions())
        squeeze_dims = data.get_slice_dimensions()
        if max_frames > 1 or pData._get_no_squeeze() or pad:
            squeeze_dims = squeeze_dims[1:]
            n_core_dims +=1
        if pData._get_rank_inc():
            sl = [(slice(None))]*n_core_dims + [None]*pData._get_rank_inc()
            return lambda x: np.squeeze(x[tuple(sl)], axis=squeeze_dims)
        return lambda x: np.squeeze(x, axis=squeeze_dims)

    def _remove_excess_data(self, data, result, slice_list):
        """ Remove any excess results due to padding for fixed length process \
        frames. """

        mData = data._get_plugin_data().meta_data.get_dictionary()
        temp = np.where(np.array(mData['size_list']) > 1)[0]
        sdir = mData['sdir'][temp[-1] if temp.size else 0]

        # Not currently working for basic_transport
        if isinstance(slice_list, slice):
            return

        sl = slice_list[sdir]
        shape = result.shape

        if shape[sdir] - (sl.stop - sl.start):
            unpad_sl = [slice(None)]*len(shape)
            unpad_sl[sdir] = slice(0, sl.stop - sl.start)
            result = result[tuple(unpad_sl)]
        return result

    def _setup_h5_files(self):
        out_data_dict = self.exp.index["out_data"]

        current_and_next = False
        if 'current_and_next' in self.exp.meta_data.get_dictionary():
            current_and_next = self.exp.meta_data.get('current_and_next')

        count = 0
        for key in out_data_dict.keys():
            out_data = out_data_dict[key]
            filename = self.exp.meta_data.get(["filename", key])
            out_data.backing_file = self.hdf5._open_backing_h5(filename, 'a')
            c_and_n = 0 if not current_and_next else current_and_next[key]
            out_data.group_name, out_data.group = self.hdf5._create_entries(
                out_data, key, c_and_n)
            count += 1

    def _set_file_details(self, files):
        self.exp.meta_data.set('link_type', files['link_type'])
        self.exp.meta_data.set('link_type', {})
        self.exp.meta_data.set('filename', {})
        self.exp.meta_data.set('group_name', {})
        for key in list(self.exp.index['out_data'].keys()):
            self.exp.meta_data.set(['link_type', key], files['link_type'][key])
            self.exp.meta_data.set(['filename', key], files['filename'][key])
            self.exp.meta_data.set(['group_name', key],
                                   files['group_name'][key])

    def _get_filenames(self, plugin_dict):
        count = self.exp.meta_data.get('nPlugin') + 1
        files = {"filename": {}, "group_name": {}, "link_type": {}}
        for key in list(self.exp.index["out_data"].keys()):
            name = key + '_p' + str(count) + '_' + \
                plugin_dict['id'].split('.')[-1] + '.h5'
            link_type = self._get_link_type(key)
            files['link_type'][key] = link_type
            if link_type == 'final_result':
                out_path = self.exp.meta_data.get('out_path')
            else:
                out_path = self.exp.meta_data.get('inter_path')

            filename = os.path.join(out_path, name)
            group_name = "%i-%s-%s" % (count, plugin_dict['name'], key)
            files["filename"][key] = filename
            files["group_name"][key] = group_name

        return files

    def _get_link_type(self, name):
        idx = self.exp.meta_data.get('nPlugin')
        temp = [e for entry in self.data_flow[idx+1:] for e in entry]
        if name in temp or self.exp.index['out_data'][name].remove:
            return 'intermediate'
        return 'final_result'

    def _populate_nexus_file(self, data):
        filename = self.exp.meta_data.get('nxs_filename')

        with h5py.File(filename, 'a') as nxs_file:
            nxs_entry = nxs_file['entry']
            name = data.data_info.get('name')
            group_name = self.exp.meta_data.get(['group_name', name])
            link_type = self.exp.meta_data.get(['link_type', name])

            if link_type is 'final_result':
                group_name = 'final_result_' + data.get_name()
            else:
                link = nxs_entry.require_group(link_type.encode("ascii"))
                link.attrs[NX_CLASS] = 'NXcollection'
                nxs_entry = link

            # delete the group if it already exists
            if group_name in nxs_entry:
                del nxs_entry[group_name]

            plugin_entry = nxs_entry.require_group(group_name)
            plugin_entry.attrs[NX_CLASS] = 'NXdata'
            self._output_metadata(data, plugin_entry, name)

    def _output_metadata(self, data, entry, name, dump=False):
        self.__output_data_type(entry, data, name)
        mDict = data.meta_data.get_dictionary()
        self._output_metadata_dict(entry.require_group('meta_data'), mDict)

        if not dump:
            self.__output_axis_labels(data, entry)
            self.__output_data_patterns(data, entry)
            if self.exp.meta_data.get('link_type')[name] == 'input_data':
                # output the filename
                entry['file_path'] = \
                    os.path.abspath(self.exp.meta_data.get('data_file'))

    def __output_data_type(self, entry, data, name):
        data = data.data if 'data' in list(data.__dict__.keys()) else data
        if isinstance(data, h5py.Dataset):
            return

        entry = entry.require_group('data_type')
        entry.attrs[NX_CLASS] = 'NXcollection'

        ltype = self.exp.meta_data.get('link_type')
        if name in list(ltype.keys()) and ltype[name] == 'input_data':
            self.__output_data(entry, data.__class__.__name__, 'cls')
            return

        args, kwargs, cls, extras = data._get_parameters(data.get_clone_args())

        for key, value in kwargs.items():
            gp = entry.require_group('kwargs')
            if isinstance(value, BaseType):
                self.__output_data_type(gp.require_group(key), value, key)
            else:
                self.__output_data(gp, value, key)

        for key, value in extras.items():
            gp = entry.require_group('extras')
            if isinstance(value, BaseType):
                self.__output_data_type(gp.require_group(key), value, key)
            else:
                self.__output_data(gp, value, key)

        for i in range(len(args)):
            gp = entry.require_group('args')
            self.__output_data(gp, args[i], ''.join(['args', str(i)]))

        self.__output_data(entry, cls, 'cls')

        if 'data' in list(data.__dict__.keys()) and not \
                isinstance(data.data, h5py.Dataset):
            gp = entry.require_group('data')
            self.__output_data_type(gp, data.data, 'data')

    def __output_data(self, entry, data, name):
        if isinstance(data, dict):
            entry = entry.require_group(name)
            entry.attrs[NX_CLASS] = 'NXcollection'
            for key, value in data.items():
                self.__output_data(entry, value, key)
        else:
            try:
                self.__create_dataset(entry, name, data)
            except Exception:
                try:
                    import json
                    data = np.array([json.dumps(data).encode("ascii")])
                    self.__create_dataset(entry, name, data)
                except Exception:
                    try:
                        data = cu._savu_encoder(data)
                        self.__create_dataset(entry, name, data)
                    except:
                        raise Exception('Unable to output %s to file.' % name)

    def __create_dataset(self, entry, name, data):
        if name not in list(entry.keys()):
            entry.create_dataset(name, data=data)
        else:
            entry[name][...] = data

    def __output_axis_labels(self, data, entry):
        axis_labels = data.data_info.get("axis_labels")
        ddict = data.meta_data.get_dictionary()

        axes = []
        count = 0
        for labels in axis_labels:
            name = list(labels.keys())[0]
            axes.append(name)
            entry.attrs[name + '_indices'] = count

            mData = ddict[name] if name in list(ddict.keys()) \
                else np.arange(data.get_shape()[count])
            if isinstance(mData, list):
                mData = np.array(mData)

            if 'U' in str(mData.dtype):
                mData = mData.astype(np.string_)

            axis_entry = entry.require_dataset(name, mData.shape, mData.dtype)
            axis_entry[...] = mData[...]
            axis_entry.attrs['units'] = list(labels.values())[0]
            count += 1
        entry.attrs['axes'] = axes

    def __output_data_patterns(self, data, entry):
        data_patterns = data.data_info.get("data_patterns")
        entry = entry.require_group('patterns')
        entry.attrs[NX_CLASS] = 'NXcollection'
        for pattern in data_patterns:
            nx_data = entry.require_group(pattern)
            nx_data.attrs[NX_CLASS] = 'NXparameters'
            values = data_patterns[pattern]
            self.__output_data(nx_data, values['core_dims'], 'core_dims')
            self.__output_data(nx_data, values['slice_dims'], 'slice_dims')

    def _output_metadata_dict(self, entry, mData):
        entry.attrs[NX_CLASS] = 'NXcollection'
        for key, value in mData.items():
            nx_data = entry.require_group(key)
            if isinstance(value, dict):
                self._output_metadata_dict(nx_data, value)
            else:
                nx_data.attrs[NX_CLASS] = 'NXdata'
                self.__output_data(nx_data, value, key)



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: base_loader
   :platform: Unix
   :synopsis: A base class for loader plugins

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""

from savu.plugins.plugin import Plugin


class BaseLoader(Plugin):
    """
    A base plugin from which all data loader plugins should inherit.

    :u*param preview: A slice list of required frames. Default: [].
    :*param data_file: Hidden parameter for Savu template. Default: "<>".
    :~param in_datasets:
    :~param out_datasets:

    """

    def __init__(self, name="baseLoader"):
        super(BaseLoader, self).__init__(name)
        self.hits = []
        self.application = None
        self.reduction_flag = False

    def _main_setup(self):
        """
        Overwrites the main_setup function in plugin.py as the loader is a
        special case of plugin that doesn't require setup of in/out_datasets
        """
        data_obj = self.setup()
        if data_obj and not self.reduction_flag:
            self.set_data_reduction_params(data_obj)

    def get_experiment(self):
        return self.exp

    def set_data_reduction_params(self, data_obj):
        pDict = self.parameters
        self.data_mapping()
        data_obj.get_preview().set_preview(pDict['preview'], load=True)
        # update axis labels
        #data_obj.amend_axis_label_values()
        # set previewing for Related datasets
        #for data in data_obj.related.values():
            # find common axes
        #    pass
        # update axis labels for Related datasets
        self.reduction_flag = True

    def get_NXapp(self, ltype, nx_file, entry):
        '''
        finds an application definition in a nexus file
        '''
        self.application = ltype
        nx_file[entry].visititems(self._visit_NXapp)
        return self.hits

    def _visit_NXapp(self, name, obj):
        if "NX_class" in list(obj.attrs.keys()):
            if obj.attrs["NX_class"].decode("ascii") in ["NXentry", "NXsubentry"]:
                if "definition" in list(obj.keys()):
                    if obj["definition"][()].decode("ascii") == self.application:
                        self.hits.append(obj)

    def get_NXdata(self, nx_file, detector_list):
        nx_file['/'].visititems(self._visit_NXdata)
        for detector in detector_list:
            for nxdata in self.nxdata:
                if detector in list(nxdata.keys()) or detector in \
                        str(nxdata.name).split('/'):
                    self.hits.append(nxdata)
        return self.hits

    def _visit_NXdata(self, name, obj):
        if "NX_class" in list(obj.attrs.keys()):
            if obj.attrs["NX_class"].decode("ascii") in ["NXdata"]:
                self.hits.append(obj)

    def data_mapping(self):
        pass



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: astra_recon_gpu
   :platform: Unix
   :synopsis: Wrapper around the Astra toolbox for gpu reconstruction using vector geometry
.. moduleauthor:: Mark Basham <scientificsoftware@diamond.ac.uk>

"""
import astra
import numpy as np

from savu.plugins.reconstructions.astra_recons.base_astra_vector_recon \
    import BaseAstraVectorRecon
from savu.plugins.driver.gpu_plugin import GpuPlugin
from savu.data.plugin_list import CitationInformation
from savu.plugins.utils import register_plugin


@register_plugin
class AstraReconGpu(BaseAstraVectorRecon, GpuPlugin):
    """
    A Plugin to run the astra reconstruction (GPU) for vector geometry

    :u*param res_norm: Output the residual norm at each iteration\
        (Error in the solution - iterative solvers only). Default: False.
    :u*param algorithm: Reconstruction type (FBP_CUDA|SIRT_CUDA|\
        SART_CUDA (not currently working)|CGLS_CUDA|FP_CUDA|BP_CUDA|\
        SIRT3D_CUDA|CGLS3D_CUDA). Default: 'FBP_CUDA'.
    :u*param FBP_filter: The FBP reconstruction filter type (none|ram-lak|\
        shepp-logan|cosine|hamming|hann|tukey|lanczos|triangular|gaussian|\
        barlett-hann|blackman|nuttall|blackman-harris|blackman-nuttall|\
        flat-top|kaiser|parzen). Default: 'ram-lak'.
    """

    def __init__(self):
        super(AstraReconGpu, self).__init__("AstraReconGpu")
        self.GPU_index = None
        self.res = False
        self.start = 0

    def set_options(self, cfg):
        if 'option' not in cfg.keys():
            cfg['option'] = {}
        cfg['option']['GPUindex'] = self.parameters['GPU_index']
        return cfg

    def nOutput_datasets(self):
        alg = self.parameters['algorithm']
        if self.parameters['res_norm'] is True and 'FBP' not in alg:
            self.res = True
            self.parameters['out_datasets'].append('res_norm')
            return 2
        return 1

    def astra_setup(self):
        options_list = ["FBP_CUDA", "SIRT_CUDA", "SART_CUDA", "CGLS_CUDA",
                        "FP_CUDA", "BP_CUDA", "SIRT3D_CUDA", "CGLS3D_CUDA"]
        if not options_list.count(self.parameters['algorithm']):
            raise Exception("Unknown Astra GPU algorithm.")

    def astra_2D_vector_recon(self, data):
        sino = data[0]
        cor, angles, vol_shape, init = self.get_frame_params()
        angles = np.deg2rad(angles)
        if self.res:
            res = np.zeros(self.len_res)
        # create volume geom
        vol_geom = astra.create_vol_geom(vol_shape)
        # create projection geom
        det_width = sino.shape[self.dim_detX]
        half_det_width = 0.5*det_width
        cor_astra = half_det_width - cor
        # set parallel beam vector geometry
        vectors = self.vec_geom_init2D(angles, 1.0, cor_astra-0.5)
        try:
            #vector geometry (astra > 1.9v)
            proj_geom = astra.create_proj_geom('parallel_vec', det_width, vectors)
        except:
            print('Warning: using scalar geometry since the Astra version <1.9 does not support the vector one for 2D')
            proj_geom = astra.create_proj_geom('parallel', 1.0, det_width, angles)
        sino = np.transpose(sino, (self.dim_rot, self.dim_detX))

        # Create a data object to hold the sinogram data
        sino_id = astra.data2d.create('-sino', proj_geom, sino)

        # create reconstruction id
        if init is not None:
            rec_id = astra.data2d.create('-vol', vol_geom, init)
        else:
            rec_id = astra.data2d.create('-vol', vol_geom)

#        if self.mask_id:
#            self.mask_id = astra.data2d.create('-vol', vol_geom, self.mask)
        # setup configuration options
        cfg = self.set_config(rec_id, sino_id, proj_geom, vol_geom)
        # create algorithm id
        alg_id = astra.algorithm.create(cfg)
        # run algorithm
        if self.res:
            for j in range(self.iters):
                # Run a single iteration
                astra.algorithm.run(alg_id, 1)
                res[j] = astra.algorithm.get_res_norm(alg_id)
        else:
            astra.algorithm.run(alg_id, self.iters)
        # get reconstruction matrix

        if self.manual_mask is not False:
            recon = self.manual_mask*astra.data2d.get(rec_id)
        else:
            recon = astra.data2d.get(rec_id)

        # delete geometry
        self.delete(alg_id, sino_id, rec_id, False)
        return [recon, res] if self.res else recon

    def astra_3D_vector_recon(self, data):
        #def astra_3D_vector_recon(self, sino, cors, angles, vol_shape, init):
#        while len(cors) is not self.sino_shape[self.slice_dir]:
#            cors.append(0)
        sino = data[0]
        cor, angles, vol_shape, init = self.get_frame_params()
        half_det_width = 0.5*sino.shape[self.sino_dim_detX]
        cor_astra = half_det_width - cor[0]

        sslice = [slice(None)]*self.nDims
        recon = np.zeros(vol_shape)
        recon = np.expand_dims(recon, axis=self.slice_dir)
        if self.res:
            res = np.zeros((self.vol_shape[self.slice_dir], self.iters))

        # create volume geometry
        vol_geom = \
            astra.create_vol_geom(vol_shape[0], vol_shape[2], vol_shape[1])
        # pad the sinogram
        # Don't pad the sinogram if 3d
        # Originally in pad_sino:
        # centre_pad = (0, 0) if '3D' in self.alg else \
        # self.array_pad(cor, sino.shape[self.dim_detX])

        #pad_sino = self.pad_sino(self.slice_func(sino, sslice), cors)
        #nDets = pad_sino.shape[self.slice_dir]
        #trans = (self.slice_dir, self.det_rot, self.sino_dim_detX)
        #pad_sino = np.transpose(pad_sino, trans)

        # create projection geom
        #vectors3D = self.create_3d_vector_geom(angles, cor, sino.shape[self.sino_dim_detX])
        # define astra vector geometry (default)
        vectors3D = self.vec_geom_init3D(np.deg2rad(angles), 1.0, 1.0, cor_astra-0.5)
        proj_geom = astra.create_proj_geom('parallel3d_vec', sino.shape[self.sino_dim_detY], sino.shape[self.sino_dim_detX],
                                           vectors3D)
        # create sinogram id
        sino_id = astra.data3d.create("-sino", proj_geom, np.swapaxes(sino, 0, 1))

        # create reconstruction id
        if init is not None:
            #init = np.transpose(init, (0, 2, 1))  # make this general
            rec_id = astra.data3d.create('-vol', vol_geom, init)
        else:
            rec_id = astra.data3d.create('-vol', vol_geom)

        # setup configuration options
        cfg = self.set_config(rec_id, sino_id, proj_geom, vol_geom)

        # create algorithm id
        alg_id = astra.algorithm.create(cfg)

        # run algorithm
        """
        if self.res:
            for j in range(self.iters):
                # Run a single iteration
                astra.algorithm.run(alg_id, 1)
                res[j] = astra.algorithm.get_res_norm(alg_id)
        else:
            astra.algorithm.run(alg_id, self.iters)
        """
        astra.algorithm.run(alg_id, self.iters)

        # get reconstruction matrix
        #if self.manual_mask:
        #    recon = self.mask*astra.data3d.get(rec_id)
        #else:
        #    recon = astra.data3d.get(rec_id)

        recon = np.transpose(astra.data3d.get(rec_id), (2, 0, 1))

        # delete geometry
        self.delete(alg_id, sino_id, rec_id, False)

        self.start += 1
        if self.res:
            return [recon, res]
        else:
            return recon

    def create_3d_vector_geom(self, angles, cors, detX):
        # add tilt for detector
        # make sure output volume is the correct way
        # add res_norm
        # add a mask
        angles = np.deg2rad(angles)
        vectors3D = np.zeros((len(angles), 12))
        shift = detX/2.0 - cors[0] # temporary
        for i in range(len(angles)):
            # ray direction
            vectors3D[i, 0] = np.cos(angles[i])
            vectors3D[i, 1] = -np.sin(angles[i])
            vectors3D[i, 2] = 0

            # center of detector
            vectors3D[i, 3] = -shift*np.sin(angles[i])
            vectors3D[i, 4] = -shift*np.cos(angles[i])
            vectors3D[i, 5] = 0

            # vector from detector pixel (0,0) to (0,1)
            vectors3D[i, 6] = -np.sin(angles[i])
            vectors3D[i, 7] = -np.cos(angles[i])
            vectors3D[i, 8] = 0

            # vector from detector pixel (0,0) to (1,0)
            vectors3D[i, 9] = 0
            vectors3D[i, 10] = 0
            vectors3D[i, 11] = 1
        return vectors3D

    def rotation_matrix2D(self, theta):
        #define 2D rotation matrix
        return np.array([[np.cos(theta), -np.sin(theta)],
                         [np.sin(theta), np.cos(theta)]])

    def rotation_matrix3D(self, theta):
        #define 3D rotation matrix
        return np.array([[np.cos(theta), -np.sin(theta), 0.0],
                         [np.sin(theta), np.cos(theta), 0.0],
                         [0.0 , 0.0 , 1.0]])

    def vec_geom_init2D(self, angles_rad, DetectorSpacingX, CenterRotOffset):
        #define 2D vector geometry
        s0 = [0.0, -1.0] # source
        d0 = [CenterRotOffset, 0.0] # detector
        u0 = [DetectorSpacingX, 0.0] # detector coordinates

        vectors = np.zeros([angles_rad.size,6])
        for i in range(0,angles_rad.size):
            theta = angles_rad[i]
            vec_temp = np.dot(self.rotation_matrix2D(theta),s0)
            vectors[i,0:2] = vec_temp[:] # ray position
            vec_temp = np.dot(self.rotation_matrix2D(theta),d0)
            vectors[i,2:4] = vec_temp[:] # center of detector position
            vec_temp = np.dot(self.rotation_matrix2D(theta),u0)
            vectors[i,4:6] = vec_temp[:] # detector pixel (0,0) to (0,1).
        return vectors

    def vec_geom_init3D(self, angles_rad, DetectorSpacingX, DetectorSpacingY, CenterRotOffset):
        #define 3D vector geometry
        s0 = [0.0, -1.0, 0.0] # source
        d0 = [CenterRotOffset, 0.0, 0.0] # detector
        u0 = [DetectorSpacingX, 0.0, 0.0] # detector coordinates
        v0 = [0.0, 0.0, DetectorSpacingY] # detector coordinates

        vectors = np.zeros([angles_rad.size,12])
        for i in range(0,angles_rad.size):
            theta = angles_rad[i]
            vec_temp = np.dot(self.rotation_matrix3D(theta),s0)
            vectors[i,0:3] = vec_temp[:] # ray position
            vec_temp = np.dot(self.rotation_matrix3D(theta),d0)
            vectors[i,3:6] = vec_temp[:] # center of detector position
            vec_temp = np.dot(self.rotation_matrix3D(theta),u0)
            vectors[i,6:9] = vec_temp[:] # detector pixel (0,0) to (0,1).
            vec_temp = np.dot(self.rotation_matrix3D(theta),v0)
            vectors[i,9:12] = vec_temp[:] # Vector from detector pixel (0,0) to (1,0)
        return vectors

    def get_citation_information(self):
        cite_info = CitationInformation()
        cite_info.name = 'citation3'
        cite_info.description = \
            ("The tomography reconstruction algorithm used in this processing \
             pipeline is part of the ASTRA Toolbox")
        cite_info.bibtex = \
            ("@article{palenstijn2011performance,\n" +
             "title={Performance improvements for iterative electron \
             tomography reconstruction using graphics processing units \
             (GPUs)},\n" +
             "author={Palenstijn, WJ and Batenburg, KJ and Sijbers, J},\n" +
             "journal={Journal of structural biology},\n" +
             "volume={176},\n" +
             "number={2},\n" +
             "pages={250--253},\n" +
             "year={2011},\n" +
             "publisher={Elsevier}\n" +
             "}")
        cite_info.endnote = \
            ("%0 Journal Article\n" +
             "%T Performance improvements for iterative electron tomography \
             reconstruction using graphics processing units (GPUs)\n" +
             "%A Palenstijn, WJ\n" +
             "%A Batenburg, KJ\n" +
             "%A Sijbers, J\n" +
             "%J Journal of structural biology\n" +
             "%V 176\n" +
             "%N 2\n" +
             "%P 250-253\n" +
             "%@ 1047-8477\n" +
             "%D 2011\n" +
             "%I Elsevier\n")
        cite_info.doi = "doi: 10.1016/j.jsb.2011.07.017"

        return super(AstraReconGpu, self).get_citation_information() + \
            [cite_info]



# Copyright 2015 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

'''
Created on 24 Nov 2017

@author: qmm55171
'''

import argparse

import savu.plugins.loaders.utils.yaml_utils as yu


def arg_parser():
    """ Arg parser for command line arguments.
    """
    parser = argparse.ArgumentParser(prog='update_template')
    parser.add_argument('template', help='Input template file.')
    parser.add_argument('entry', help='the plugin entry')
    parser.add_argument('key', help='The parameter key.')
    parser.add_argument('value', help='The parameter value.')
    return parser.parse_args()


def update_template(filename, entry, key, val):
    template = yu.read_yaml(filename)
    entry = _string_to_val(entry)
    val = _string_to_val(val)
    template[entry][list(template[entry].keys())[0]][key] = val

    with open(filename, 'w') as stream:
        yu.dump_yaml(template, stream)


def _string_to_val(string):
    if string.isdigit():
        return (int(string))
    try:
        return float(string)
    except ValueError:
        return string

if __name__ == '__main__':
    arg = arg_parser()
    update_template(arg.template, arg.entry, arg.key, arg.value)



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: time_based_correction
   :platform: Unix
   :synopsis: A time-based dark and flat field correction using linear\
       interpolation

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""

import numpy as np

from savu.plugins.driver.cpu_plugin import CpuPlugin
from savu.plugins.corrections.base_correction import BaseCorrection
from savu.plugins.utils import register_plugin


@register_plugin
class TimeBasedCorrection(BaseCorrection, CpuPlugin):
    """
    Apply a time-based dark and flat field correction to data.

    :param in_range: Set to True if you require values in the \
        range [0, 1]. Default: False.
    """

    def __init__(self, name="TimeBasedCorrection"):
        super(TimeBasedCorrection, self).__init__(name)

    def pre_process(self):
        self.count = 0
        inData = self.get_in_datasets()[0]
        pData = self.get_plugin_in_datasets()[0]
        self.mfp = inData._get_plugin_data()._get_max_frames_process()
        self.proj_dim = \
            inData.get_data_dimension_by_axis_label('rotation_angle')
        self.slice_dir = pData.get_slice_dimension()
        nDims = len(pData.get_shape())
        self.sslice = [slice(None)]*nDims

        self.image_key = inData.data.get_image_key()
        changes = np.where(np.diff(self.image_key) != 0)[0] + 1
        self.split_key = np.split(self.image_key, changes)
        self.split_idx = np.split(np.arange(len(self.image_key)), changes)
        self.data_key = inData.data.get_index(0)

        self.dark, self.dark_idx = self.calc_average(inData.data.dark(), 2)
        self.flat, self.flat_idx = self.calc_average(inData.data.flat(), 1)

        inData.meta_data.set('multiple_dark', self.dark)
        inData.meta_data.set('multiple_flat', self.flat)

    def calc_average(self, data, key):
        im_key = np.where(self.image_key == key)[0]
        splits = np.where(np.diff(im_key) > 1)[0]+1
        local_idx = np.split(np.arange(len(im_key)), splits)
        mean_data = [np.mean(data[np.array(local_idx[i])], axis=0)
                     for i in range(len(local_idx))]
        list_idx = list(np.where([key in i for i in self.split_key])[0])
        return mean_data, list_idx

    def process_frames(self, data):
        proj = data[0]
        frame = self.get_global_frame_index()[self.count]
        flat = self.calculate_flat_field(
                *self.find_nearest_frames(self.flat_idx, frame))
        dark = self.calculate_dark_field(
                *self.find_nearest_frames(self.dark_idx, frame))

        if self.parameters['in_range']:
            proj = self.in_range(proj, flat)

        self.count += 1
        return np.nan_to_num((proj-dark)/(flat-dark))

    def in_range(self, data, flat):
        data[data > flat] = flat[data > flat]
        return data

    def find_nearest_frames(self, idx_list, value):
        """ Find the index of the two entries that 'value' lies between in \
            'idx_list' and calculate the distance between each of them.
        """
        global_val = self.data_key[value]
        # find which list (index) global_val belongs to
        list_idx = [global_val in i for i in self.split_idx].index(True)
        val_list = self.split_idx[list_idx]
        # find length of list
        length_list = len(val_list)
        # find position of global_val in list and distance from each end
        pos = np.where(val_list == global_val)[0][0]
        dist = [(length_list-pos)/float(length_list), pos/float(length_list)]

        # find closest before and after idx_list entries
        new_list = list(np.sort(np.append(idx_list, list_idx)))
        new_idx = new_list.index(list_idx)

        entry1 = new_idx-1 if new_idx != 0 else new_idx+1
        entry2 = new_idx+1 if new_idx != len(new_list)-1 else new_idx-1
        before = idx_list.index(new_list[entry1])
        after = idx_list.index(new_list[entry2])

        return [before, after], dist

    def calculate_flat_field(self, frames, distance):
        return self.flat[frames[0]]*distance[0] + \
            self.flat[frames[1]]*distance[1]

    def calculate_dark_field(self, frames, distance):
        return self.dark[frames[0]]*distance[0] + \
            self.dark[frames[1]]*distance[1]

    def get_max_frames(self):
        return 'single'



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: scikitimage_filter_back_projection
   :platform: Unix
   :synopsis: Wrapper for scikitimage FBP function

"""

import logging
from savu.plugins.reconstructions.base_recon import BaseRecon
from savu.data.plugin_list import CitationInformation
from savu.plugins.driver.cpu_plugin import CpuPlugin

import skimage.transform as transform
import numpy as np
from scipy import ndimage

from savu.plugins.utils import register_plugin


@register_plugin
class ScikitimageFilterBackProjection(BaseRecon, CpuPlugin):
    """
    A Plugin to reconstruct an image by filter back projection
    using the inverse radon transform from scikit-image.

    :param output_size: Number of rows and columns in the \
        reconstruction. Default: 'auto'.
    :param filter: Filter used in frequency domain filtering Ramp filter used \
        by default. Filters available: ramp, shepp-logan, cosine, hamming, \
        hann. Assign None to use no filter. Default: 'ramp'.
    :param interpolation: interpolation method used in reconstruction. \
        Methods available: 'linear', 'nearest', and 'cubic' \
        ('cubic' is slow). Default: 'linear'.
    :param circle: Assume the reconstructed image is zero outside the \
        inscribed circle. Also changes the default output_size to match the \
        behaviour of radon called with circle=True. Default: False.

    :~param outer_pad: Not required. Default: False.
    :~param centre_pad: Not required. Default: False.
    """

    def __init__(self):
        logging.debug("initialising Scikitimage Filter Back Projection")
        logging.debug("Calling super to make sure that all superclasses are " +
                      " initialised")
        super(ScikitimageFilterBackProjection,
              self).__init__("ScikitimageFilterBackProjection")

    def _shift(self, sinogram, centre_of_rotation):
        centre_of_rotation_shift = (sinogram.shape[0] // 2) - centre_of_rotation
        result = ndimage.interpolation.shift(sinogram,
                                             (centre_of_rotation_shift, 0))
        return result

    def process_frames(self, data):
        sino = data[0]
        centre_of_rotations, angles, vol_shape, init = self.get_frame_params()
        in_pData = self.get_plugin_in_datasets()[0]
        in_meta_data = self.get_in_meta_data()[0]
        sinogram = np.swapaxes(sino, 0, 1)
        sinogram = self._shift(sinogram, centre_of_rotations)
        theta = in_meta_data.get('rotation_angle')

        dim_detX = in_pData.get_data_dimension_by_axis_label('detector_x')
        size = self.parameters['output_size']
        size = in_pData.get_shape()[dim_detX] if size == 'auto' or \
            size is None else size

        result = \
            transform.iradon(sinogram, theta=theta,
                             output_size=(size),
                             filter=self.parameters['filter'],
                             interpolation=self.parameters['interpolation'],
                             circle=self.parameters['circle'])
        return result

    def get_max_frames(self):
        return 'single'

    def get_citation_information(self):
        cite_info = CitationInformation()
        cite_info.description = \
            ("The Tomographic reconstruction performed in this processing " +
             "chain is derived from this work.")
        cite_info.bibtex = \
            ("@book{avinash2001principles,\n" +
             " title={Principles of computerized tomographic imaging},\n" +
             " author={Kak, Avinash C. and Slaney, Malcolm},\n" +
             " year={2001},\n" +
             " publisher={Society for Industrial and Applied Mathematics}\n" +
             "}")
        cite_info.endnote = \
            ("%0 Book\n" +
             "%T Principles of computerized tomographic imaging\n" +
             "%A Kak, Avinash C.\n" +
             "%A Slaney, Malcolm\n" +
             "%@ 089871494X\n" +
             "%D 2001\n" +
             "%I Society for Industrial and Applied Mathematics")
        cite_info.doi = "https://doi.org/10.1137/1.9780898719277"
        return cite_info



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: Remove stripe artefacts
   :platform: Unix
   :synopsis: A plugin working in sinogram space to remove stripe artefacts
.. moduleauthor:: Nghia Vo <scientificsoftware@diamond.ac.uk>

"""

from savu.plugins.plugin import Plugin
from savu.plugins.driver.cpu_plugin import CpuPlugin
from savu.plugins.utils import register_plugin
from savu.data.plugin_list import CitationInformation

import numpy as np
from scipy.ndimage import median_filter
from scipy import signal
import pyfftw.interfaces.scipy_fftpack as fft

@register_plugin
class RingRemovalFiltering(Plugin, CpuPlugin):
    """

    Method to remove stripe artefacts in a sinogram (<-> ring artefacts in a \
    reconstructed image) using a filtering-based method in the combination \
    with a sorting-based method. Note that it's different to a FFT-based or \
    wavelet-FFT-based method.

    :param sigma: Sigma of the Gaussian window. Used to separate the low-pass\
     and high-pass components of each sinogram column. Default: 3.
    :param size: Size of the median filter window. Used to\
     clean stripes. Default: 31.

    """

    def __init__(self):
        super(RingRemovalFiltering, self).__init__(
                "RingRemovalFiltering")

    def setup(self):
        in_dataset, out_dataset = self.get_datasets()
        out_dataset[0].create_dataset(in_dataset[0])
        in_pData, out_pData = self.get_plugin_datasets()
        in_pData[0].plugin_data_setup('SINOGRAM', 'single')
        out_pData[0].plugin_data_setup('SINOGRAM', 'single')

    def pre_process(self):
        in_pData = self.get_plugin_in_datasets()
        width_dim = \
            in_pData[0].get_data_dimension_by_axis_label('detector_x')
        height_dim = \
            in_pData[0].get_data_dimension_by_axis_label('rotation_angle')
        sino_shape = list(in_pData[0].get_shape())
        self.pad = 150 # To reduce artifact caused by FFT
        self.width1 = sino_shape[width_dim] 
        self.height1 = sino_shape[height_dim] + 2*self.pad
        listindex = np.arange(0.0, sino_shape[height_dim], 1.0)
        self.matindex = np.tile(listindex, (self.width1, 1))        
        sigma = np.clip(np.int16(self.parameters['sigma']), 1, self.height1-1)
        self.window = signal.gaussian(self.height1, std = sigma)
        self.listsign = np.power(-1.0,np.arange(self.height1))

    def remove_stripe_based_sorting(self, matindex, sinogram, size):
        """
        Remove stripes using the sorting technique.
        ---------
        Parameters: - sinogram: 2D array.
                    - size: window size of the median filter.
        ---------
        Return:     - stripe-removed sinogram.
        """
        sinogram = np.transpose(sinogram)
        matcomb = np.asarray(np.dstack((matindex, sinogram)))
        matsort = np.asarray(
            [row[row[:, 1].argsort()] for row in matcomb])
        matsort[:, :, 1] = median_filter(matsort[:, :, 1], (size, 1))
        matsortback = np.asarray(
            [row[row[:, 0].argsort()] for row in matsort])
        sino_corrected = matsortback[:, :, 1]
        return np.transpose(sino_corrected)

    def process_frames(self, data):
        sinogram = np.transpose(np.copy(data[0]))
        sinogram2 = np.pad(
            sinogram,((0, 0),(self.pad, self.pad)), mode = 'reflect')
        size = np.clip(np.int16(self.parameters['size']), 1, self.width1-1)
        sinosmooth = np.zeros_like(sinogram)
        for i,sinolist in enumerate(sinogram2):        
            sinosmooth[i] = np.real(fft.ifft(fft.fft(sinolist
                            *self.listsign)*self.window)
                            *self.listsign)[self.pad:self.height1-self.pad]
        sinosharp = sinogram - sinosmooth
        sinosmooth_cor = np.transpose(
            self.remove_stripe_based_sorting(
                self.matindex, np.transpose(sinosmooth), size))
        return np.transpose(sinosmooth_cor + sinosharp)

    def get_citation_information(self):
        cite_info = CitationInformation()
        cite_info.description = \
            ("The code of ring removal is the implementation of the work of \
            Nghia T. Vo et al. taken from algorithm 2 and 3 in this paper.")
        cite_info.bibtex = \
            ("@article{Vo:18,\n" +
             "title={Superior techniques for eliminating ring artifacts in\
              X-ray micro-tomography},\n" +
             "author={Nghia T. Vo, Robert C. Atwood,\
              and Michael Drakopoulos},\n" +
             "journal={Opt. Express},\n" +
             "volume={26},\n" +
             "number={22},\n" +
             "pages={28396--28412},\n" +
             "year={2018},\n" +
             "publisher={OSA}" +
             "}")
        cite_info.doi = "doi: DOI: 10.1364/OE.26.028396"
        return cite_info



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: tomo_recon
   :platform: Unix
   :synopsis: runner for tests using the MPI framework

.. moduleauthor:: Mark Basham <scientificsoftware@diamond.ac.uk>

"""
import unittest
from savu.test import test_utils as tu
from savu.test.travis.framework_tests.plugin_runner_test import \
    run_protected_plugin_runner


class PluginRunnerScikitImageTest(unittest.TestCase):
    global data_file, experiment
    data_file = '24737.nxs'
    experiment = None

    def test_process(self):
        process_list = 'reconstruction/scikit_sart_test_process.nxs'
        options = tu.initialise_options(data_file, experiment, process_list)
        run_protected_plugin_runner(options)
        tu.cleanup(options)

if __name__ == "__main__":
    unittest.main()



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: framework_citations
   :platform: Unix
   :synopsis: Contains all framework citations.

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""


def get_framework_citations():
    """ return a list of NXcite objects """

    cite_info = []

    cite_info.append({'name': 'Savu'})
    cite_info[0]['description'] = \
        ("The Savu framework design is described in this paper.")
    cite_info[0]['bibtex'] = \
        ("@article{wadeson2016savu,\n" +
         "title={Savu: A Python-based, MPI Framework for Simultaneous\n" +
         "Processing of Multiple, N-dimensional, Large Tomography\n" +
         "Datasets},\n" +
         "author={Wadeson, Nicola and Basham, Mark},\n" +
         "journal={arXiv preprint arXiv:1610.08015},\n" +
         "year={2016}\n" +
         "}")
    cite_info[0]['endnote'] = \
        ("%0 Journal Article\n" +
         "%T Savu: A Python-based, MPI Framework for Simultaneous Processing\n" +
         "of Multiple, N-dimensional, Large Tomography Datasets\n" +
         "%A Wadeson, Nicola\n" +
         "%A Basham, Mark\n" +
         "%J arXiv preprint arXiv:1610.08015\n" +
         "%D 2016")

    cite_info.append({'name': 'HDF5'})
    cite_info[1]['description'] = \
        ("Savu uses parallel HDF5 as a backend.")
    cite_info[1]['bibtex'] = \
        ("@ONLINE{hdf5, \n" +
         "author = {{The HDF Group}}, \n" +
         "title = '{Hierarchical Data Format, version 5}', \n" +
         "year = {1997-2016}, \n" +
         "note = {/HDF5/}, \n" +
         "}")
    cite_info[1]['endnote'] = \
        ("%0 Journal Article\n" +
         "%T {Hierarchical Data Format, version 5}\n" +
         "%A HDF Group\n" +
         "%D 2014\n")

    cite_info.append({'name': 'MPI'})
    cite_info[2]['description'] = ("HDF5 uses the Message Passing Interface\n" +
        "standard for interprocess communication")
    cite_info[2]['bibtex'] = \
        ("@article{walker1996mpi, \n" +
         "title={MPI: a standard message passing interface}, \n" +
         "author={Walker, David W and Dongarra, Jack J}, \n" +
         "journal={Supercomputer}, \n" +
         "volume={12}, \n" +
         "pages={56--68}, \n" +
         "year={1996}, \n" +
         "publisher={ASFRA BV} \n" +
         "}")
    cite_info[2]['endnote'] = \
        ("%0 Journal Article\n" +
         "%T MPI: a standard message passing interface\n" +
         "%A Walker, David W\n" +
         "%A Dongarra, Jack J\n" +
         "%J Supercomputer\n" +
         "%V 12\n" +
         "%P 56-68\n" +
         "%@ 0168-7875\n" +
         "%D 1996\n" +
         "%I ASFRA BV\n")

    return cite_info



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: subpixel_shift
   :platform: Unix
   :synopsis: A plugin to apply a subpixel x-shift to images
.. moduleauthor:: Malte Storm<malte.storm@diamond.ac.uk>
"""

from savu.plugins.filters.base_filter import BaseFilter
from savu.plugins.driver.cpu_plugin import CpuPlugin
from savu.plugins.utils import register_plugin

import scipy.ndimage.interpolation as sip
import numpy as np
import skimage.transform as sktf



@register_plugin
class SubpixelShift(BaseFilter, CpuPlugin):
    """
    A plugin to apply a sub-pixel correction to images, for example to allow \
    subpixel alignment for the AstraGpu plugin.
    
    :u*param x_shift: The shift in x for the output image in pixels. Positive \
    values correspond to data being shifted towards larger indices. \
    Default: 0.0.
    :param transform_module: The module (skimage|scipy) to be used for image \
    translation. skimage corresponds to skimage.transform.SimilarityTransform \
    while scipy corresponds to scipy.ndimage.interpolation. Default: 'skimage'.

    """
    def __init__(self):
        super(SubpixelShift, self).__init__('SubpixelShift')

    def setup(self):
        in_dataset, out_dataset = self.get_datasets()
        in_pData, out_pData = self.get_plugin_datasets()
        self.det_x = in_dataset[0].\
        get_data_dimension_by_axis_label('detector_x')

        out_dataset[0].create_dataset(in_dataset[0])
        in_pData[0].plugin_data_setup('SINOGRAM', 'single')
        out_pData[0].plugin_data_setup('SINOGRAM', 'single')
        if self.parameters['transform_module'] == 'skimage':
            self.process_frames = self.process_frames_skimage
        elif self.parameters['transform_module'] == 'scipy':
            self.process_frames = self.process_frames_scipy
        else:
            raise Exception('Transform module not supported.')

    def pre_process(self):
        self.xshift = float(self.parameters['x_shift'])
        if self.parameters['transform_module'] == 'skimage':
            self.xshift *= -1
            if self.xshift >=0:
                self.pad_slice = slice(0, int(np.ceil(self.xshift)))
                self.pad_col = int(np.ceil(self.xshift))
            elif self.xshift <0:
                self.pad_slice = slice(self.det_x+\
                int(np.floor(self.xshift)), self.det_x)
                self.pad_col =self.det_x+int(np.floor(self.xshift)) -1
            self.tf = sktf.SimilarityTransform(scale = 1, rotation = 0, \
            translation = (self.xshift, 0))
    
    def process_frames_scipy(self, data):
        return sip.shift(data[0], (self.xshift, 0), mode='nearest', order=3)

    def process_frames_skimage(self, data):
        tmpdata = sktf.warp(data[0].astype(np.float64), self.tf)\
        .astype(np.float32)
        tmpdata[:,self.pad_slice] = tmpdata[:, self.pad_col][:,np.newaxis]
        return tmpdata

    def post_process(self):
        pass




# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: Remove stripe artefacts
   :platform: Unix
   :synopsis: A plugin working in sinogram space to remove stripe artefacts
.. moduleauthor:: Nghia Vo <scientificsoftware@diamond.ac.uk>

"""

from savu.plugins.plugin import Plugin
from savu.plugins.driver.cpu_plugin import CpuPlugin
from savu.plugins.utils import register_plugin
from savu.data.plugin_list import CitationInformation
import numpy as np
from scipy.ndimage import median_filter


@register_plugin
class RingRemovalSorting(Plugin, CpuPlugin):
    """

    Method to remove stripe artefacts in a sinogram (<-> ring artefacts in a \
    reconstructed image) using a sorting-based method. It works particularly well\
    for removing partial rings.

    :param size: Size of the median filter window. Greater is stronger.\
     Default: 31.

    """

    def __init__(self):
        super(RingRemovalSorting, self).__init__(
                "RingRemovalSorting")

    def setup(self):
        in_dataset, out_dataset = self.get_datasets()
        out_dataset[0].create_dataset(in_dataset[0])
        in_pData, out_pData = self.get_plugin_datasets()
        in_pData[0].plugin_data_setup('SINOGRAM', 'single')
        out_pData[0].plugin_data_setup('SINOGRAM', 'single')

    def pre_process(self):
        in_pData = self.get_plugin_in_datasets()
        width_dim = \
            in_pData[0].get_data_dimension_by_axis_label('detector_x')
        height_dim = \
            in_pData[0].get_data_dimension_by_axis_label('rotation_angle')
        sino_shape = list(in_pData[0].get_shape())
        self.width1 = sino_shape[width_dim]
        self.height1 = sino_shape[height_dim]
        listindex = np.arange(0.0, self.height1, 1.0)
        self.matindex = np.tile(listindex,(self.width1,1))        

    def process_frames(self, data):
        sinogram = np.transpose(np.copy(data[0]))
        size = np.clip(np.int16(self.parameters['size']), 1, self.width1-1)
        matcomb = np.asarray(np.dstack((self.matindex, sinogram)))
        matsort = np.asarray([row[row[:, 1].argsort()] for row in matcomb])
        matsort[:, :, 1] = median_filter(matsort[:, :, 1],(size,1))
        matsortback = np.asarray(
            [row[row[:, 0].argsort()] for row in matsort])
        sino_corrected = matsortback[:, :, 1]
        return np.transpose(sino_corrected)

    def get_citation_information(self):
        cite_info = CitationInformation()
        cite_info.description = \
            ("The code of ring removal is the implementation of the work of \
            Nghia T. Vo et al. taken from algorithm 3 in this paper.")
        cite_info.bibtex = \
            ("@article{Vo:18,\n" +
             "title={Superior techniques for eliminating ring artifacts in\
              X-ray micro-tomography},\n" +
             "author={Nghia T. Vo, Robert C. Atwood,\
              and Michael Drakopoulos},\n" +
             "journal={Opt. Express},\n" +
             "volume={26},\n" +
             "number={22},\n" +
             "pages={28396--28412},\n" +
             "year={2018},\n" +
             "publisher={OSA}" +
             "}")
        cite_info.doi = "doi: DOI: 10.1364/OE.26.028396"
        return cite_info



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: plugin_coverage_test
   :platform: Unix
   :synopsis: Test that all the plugins are

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""

import unittest
import os

from savu.data.plugin_list import PluginList
import savu.test.test_process_list_utils as tplu


class PluginCoverageTest(unittest.TestCase):

    def test_coverage(self):
        savu_base_path = \
            os.path.dirname(os.path.realpath(__file__)).split('savu')[0]

        # lists all .nxs process lists used in the tests, and all plugins
        # directly called in the tests
        [nxs_in_tests, plugins_in_tests] = \
            tplu.get_process_list(savu_base_path + '/savu/test')

        # remove data files from the list
        data_list = self.get_data_list(savu_base_path + '/test_data/data')
        nxs_in_tests = list(set(nxs_in_tests).difference(set(data_list)))

        # since now some nxs files are in the subfolders, we need to make sure
        # that we extract the basename only (without the "head")
        nxs_in_tests_mod = []
        for nxs in nxs_in_tests:
            try:
                filename_nxs = os.path.basename(nxs)
                nxs_in_tests_mod.append(filename_nxs)
            except Exception:
                print("The failed basename file:", nxs)

        # list all test process lists available in test_process_lists folder
        test_process_path = savu_base_path + 'test_data/test_process_lists'
        self.nxs_avail = tplu.get_test_process_list(test_process_path)

        # list the .nxs found in tests that are located in the
        # test_process_lists folder
        self.nxs_used = \
            list(set(nxs_in_tests).intersection(set(self.nxs_avail)))

        # which test process lists were not in the test_process_lists folder
        nxs_unused = list(set(nxs_in_tests).difference(set(self.nxs_avail)))
        print("===============================================================")
        print("\nThese .nxs test files were found inside the tests, but are "
              "not available in the test_process_lists folder:\n")
        for nxs in nxs_unused:
            print (nxs)
        print ("===============================================================")

        # get all plugins listed in self.nxs_used process lists
        tested_plugin_list = \
            self.get_test_plugin_list(self.nxs_used, test_process_path)
        tested_plugin_list += plugins_in_tests

        # list all plugins
        plugin_list = self.get_plugin_list(savu_base_path + '/savu/plugins')

        print ("===============================================================")
        print ("\nThe following plugins are not covered by the tests:\n")
        uncovered = list(set(plugin_list).difference(set(tested_plugin_list)))
        for plugin in uncovered:
            print (plugin)
        print("===============================================================")

        print("===============================================================")
        print("\nThe following process lists are redundant:\n")
        redundant = list(set(self.nxs_avail).difference(set(self.nxs_used)))
        for plugin in redundant:
            print (plugin)
        print("===============================================================")

    def test_process_lists(self):
        # check for unused process lists
        pass

    def get_data_list(self, folder):
        data_list = []
        ftypes = ['nxs', 'h5']
        exclude_dir = ['__pycache__']
        exclude_file = ['__init__.py']
        for root, dirs, files in os.walk(folder, topdown=True):
            dirs[:] = [d for d in dirs if d not in exclude_dir]
            files[:] = [fi for fi in files if fi.split('.')[-1] in ftypes]
            files[:] = [fi for fi in files if fi not in exclude_file]
            for f in files:
                data_list.append(f)
        return data_list

    def get_test_plugin_list(self, process_files, path):
        plugin_names = []
        for pfile in process_files:
            plist = PluginList()
            plist._populate_plugin_list(path + '/' + pfile)
            for p in plist.plugin_list:
                try:
                    plugin_id = p['id']
                    pList = self.add_plugin(plugin_id)
                    for p in pList:
                        plugin_names.append(p + '.py')
                except ImportError as e:
                    print("Failed to run test as libraries not available (%s),"
                          % (e) + " passing test")
        return list(set(plugin_names))

    def add_plugin(self, plugin_id):
        plugin_list = []
        plugin = self.load_plugin(plugin_id)
        pList = plugin.__class__.mro()
        for p in pList:
            module_name = p.__module__.split('.')
            if module_name[0] == 'savu':
                plugin_list.append(module_name[-1])
        return plugin_list

    def get_plugin_list(self, folder):
        plugin_list = []
        exclude_dir = ['__pycache__']
        exclude_file = ['__init__.py']
        for root, dirs, files in os.walk(folder, topdown=True):
            dirs[:] = [d for d in dirs if d not in exclude_dir]
            files[:] = [fi for fi in files if fi.split('.')[-1] == 'py']
            files[:] = [fi for fi in files if fi not in exclude_file]
            for f in files:
                plugin_list.append(f)
        return plugin_list

    def load_plugin(self, name):
        mod = __import__(name)
        components = name.split('.')
        for comp in components[1:]:
            mod = getattr(mod, comp)
        module_name = name.split('.')[-1]
        mod2class = ''.join(x.capitalize() for x in module_name.split('_'))
        clazz = getattr(mod, mod2class.split('.')[-1])()
        return clazz


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
.. module:: astra_multiple_parameter_test
   :platform: Unix
   :synopsis: unittest test classes for plugins

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""

import unittest

import savu.plugins.utils as pu
import savu.test.test_utils as tu

from collections import OrderedDict


class AstraMultipleParameterTest(unittest.TestCase):

    def plugin_setup(self):
        ppath = 'savu.plugins.reconstructions.astra_recons.astra_recon_cpu'
        plugin = pu.load_class(ppath)()
        return plugin

    def framework_options_setup(self):
        key1 = 'n_iterations'
        key2 = 'algorithm'
        key3 = 'in_datasets'
        key4 = 'out_datasets'
        params = OrderedDict({key1: '1;2;3', key2: 'FBP;CGLS', key3: 'tomo', key4: 'tomo'})

        options = tu.set_experiment('tomo')
        plugin = 'savu.plugins.reconstructions.astra_recons.astra_recon_cpu'
        tu.set_plugin_list(options, plugin, [{}, params, {}])
        return options

    def test_parameter_space_int(self):
        plugin = self.plugin_setup()
        key = 'n_iterations'
        params = {key: '1;2;3'}
        plugin._set_parameters(params)
        params = plugin.parameters[key]
        self.assertEqual(params, [1, 2, 3])
        self.assertEqual(plugin.extra_dims[0], 3)

    def test_parameter_space_str(self):
        plugin = self.plugin_setup()
        key = 'algorithm'
        params = {key: 'FBP;CGLS'}
        plugin._set_parameters(params)
        params = plugin.parameters[key]
        self.assertEqual(params, ['FBP', 'CGLS'])
        self.assertEqual(plugin.extra_dims[0], 2)

    def test_parameter_space_extra_dims(self):
        plugin = self.plugin_setup()
        key1 = 'algorithm'
        key2 = 'n_iterations'
        params = {key1: 'FBP;CGLS', key2: '1;2;3'}
        plugin._set_parameters(params)
        out_datasets = plugin.get_out_datasets()
        for data in out_datasets:
            self.assertEqual(data.extra_dims, plugin.extra_dims)

    def test_parameter_space_data_shape(self):
        options = self.framework_options_setup()
        plugin = tu.plugin_runner_load_plugin(options)

        out_dataset = plugin.get_out_datasets()[0]
        self.assertEqual((160, 135, 160, 3, 2), out_dataset.get_shape())

#    def test_parameter_space_full_run(self):
#        options = self.framework_options_setup()
#        tu.plugin_runner_real_plugin_run(options)

if __name__ == "__main__":
    unittest.main()
